from fastapi import FastAPI

from fa_common import create_app, get_settings


def test_create_app(env_path):

    app = create_app(env_path=env_path)
    assert isinstance(app, FastAPI)

    settings = get_settings()
    assert settings.PROJECT_NAME == "Fast Api Commons Test"



from humps.camel import case
from pydantic import BaseModel


def to_camel(string):
    return case(string)


class CamelModel(BaseModel):
    class Config:
        alias_generator = to_camel
        allow_population_by_field_name = True


class FileDownloadRef(CamelModel):
    name: str
    url: str
    extension: str
    size: int



import time
from collections.abc import Iterable

from gitlab import Gitlab
from gitlab.v4.objects import Project, ProjectPipeline, ProjectPipelineJob, ProjectJob
from gitlab.exceptions import GitlabGetError, GitlabCreateError, GitlabDeleteError

import yaml
import json
from typing import List, Union
from fa_common import force_async, get_current_app, get_settings
from fa_common import logger as LOG
from .utils import GitlabCIYAMLDumper
from .models import WorkflowRun, JobRun


class GitlabClient:
    """
    Singleton client for interacting with gitlab.
    Is a wrapper over the existing gitlab python client to provide specialist functions for the Job/Module
    workflow.

    Please don't use it directly, use `fa_common.workflow.utils.get_workflow_client`.
    """

    __instance = None
    gitlab: Gitlab = None

    def __new__(cls) -> "GitlabClient":
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
            app = get_current_app()
            cls.__instance.gitlab = app.gitlab  # type: ignore
        return cls.__instance

    async def project_exists(self, project_id: Union[int, str]) -> bool:
        try:
            self._get_project(project_id)
        except ValueError:
            return False
        return True

    async def _get_project(self, project_id: Union[int, str]) -> Project:
        if isinstance(project_id, int):
            try:
                project = await force_async(self.gitlab.projects.get)(project_id)
            except GitlabGetError:
                raise ValueError(f"No project found with the id {project_id}")
            return project
        else:
            return await self._get_project_by_name(project_id)

    async def _get_project_by_name(self, project_name: str):
        settings = get_settings()
        projects = await force_async(self.gitlab.projects.list)(
            search=project_name, owned=True
        )
        for proj in projects:
            group_id = proj.namespace["id"]
            if proj.name == project_name and group_id == settings.GITLAB_GROUP_ID:
                return proj
        raise ValueError(f"No project found with the name {project_name}")

    async def create_project(self, project_name: str) -> int:
        settings = get_settings()
        try:
            project = await force_async(self.gitlab.projects.create)(
                {"name": project_name, "namespace_id": settings.GITLAB_GROUP_ID}
            )
        except GitlabCreateError as err:
            LOG.warning(f"Create Error caught, retrying in 5 secs: {err}")
            time.sleep(5)
            project = await force_async(self.gitlab.projects.create)(
                {"name": project_name, "namespace_id": settings.GITLAB_GROUP_ID}
            )

        LOG.info(f"Created project: {project}")

        data = {
            "branch": "master",
            "commit_message": "First Commit",
            "actions": [
                {"action": "create", "file_path": ".gitlab-ci.yml", "content": ""}
            ],
        }
        commit = await force_async(project.commits.create)(data)
        LOG.info(f"Created commit: {commit}")

        return project

    async def create_branch(self, project_id: int, branch_name: str):
        project = await self._get_project(project_id)
        branch = await force_async(project.branches.create)(
            {"branch": branch_name, "ref": "master"}
        )
        LOG.info(f"Created branch: {branch}")
        return branch

    async def delete_branch(self, project_id: int, branch_name: str):
        project = await self._get_project(project_id)
        try:
            await force_async(project.branches.delete)(branch_name)
            LOG.info(f"Deleted branch: {branch_name}")
        except GitlabDeleteError as err:
            if str(err.response_code) == "404":
                raise ValueError(
                    f"Trying to delete branch {branch_name} that doesn't exist."
                )
            raise err

    async def get_job(
        self, project_id: Union[str, int], job_id: int, include_log: bool = False
    ) -> JobRun:
        project = await self._get_project(project_id)
        try:
            job = await force_async(project.jobs.get)(job_id)
            log = None
            if include_log:
                log = job.trace()
        except GitlabGetError:
            raise ValueError(
                f"No job found with the id {job_id} in project {project_id}"
            )

        return self.job_to_job_run(job, log)

    async def get_job_log(self, project_id: Union[str, int], job_id: int) -> bytes:
        project = await self._get_project(project_id)
        job = await force_async(project.jobs.get)(job_id)
        return job.trace()

    async def update_ci(
        self,
        project_id: Union[str, int],
        branch_name: str,
        ci_file: dict,
        update_message: str = "No message",
    ):
        project = await self._get_project(project_id)
        data = {
            "branch": branch_name,
            "commit_message": update_message,
            "actions": [
                {
                    "action": "update",
                    "file_path": ".gitlab-ci.yml",
                    "content": yaml.dump(
                        ci_file, Dumper=GitlabCIYAMLDumper, default_flow_style=False
                    ),
                }
            ],
        }
        commit = await force_async(project.commits.create)(data)
        LOG.info(f"Created CI Commit: {commit}")

        return commit.id

    @LOG.catch
    async def run_pipeline(
        self,
        project_id: Union[str, int],
        branch: str,
        variables: List[dict] = None,
        wait: bool = False,
    ) -> WorkflowRun:
        project = await self._get_project(project_id)
        pipeline = project.pipelines.create({"ref": branch, "variables": variables})
        if wait:
            while pipeline.finished_at is None:
                pipeline.refresh()
                time.sleep(2)
        jobs = pipeline.jobs.list()

        return await self.pipeline_to_workflow_run(pipeline, project, jobs)

    def job_to_job_run(self, job: ProjectJob, job_log: bytes = None) -> JobRun:
        output = None
        if hasattr(job, "artifacts_file") and job.artifacts_file is not None:
            try:
                output_bytes = None
                output_bytes = job.artifact("output/outputs.json")
                if output_bytes is not None:
                    output = json.loads(output_bytes)
                    if isinstance(output, Iterable):
                        output = List(output)
            except Exception as err:
                LOG.error(err)

        job_run = JobRun(
            id=job.id,
            status=job.status,
            started_at=job.started_at,
            finished_at=job.finished_at,
            duration=job.duration,
            name=job.name,
            stage=job.stage,
            output=output,
        )
        if job_log is not None:
            job_run.log = job_log

        return job_run

    async def pipeline_to_workflow_run(
        self,
        pipeline: ProjectPipeline,
        project: Project,
        jobs: List[ProjectPipelineJob] = [],
    ) -> WorkflowRun:
        job_runs: List[JobRun] = []
        for job in jobs:
            job_runs.append(await self.get_job(project.id, job.id))

        return WorkflowRun(
            id=pipeline.id,
            gitlab_project_id=project.id,
            gitlab_project_branch=pipeline.ref,
            commit_id=pipeline.sha,
            status=pipeline.status,
            started_at=pipeline.started_at,
            finished_at=pipeline.finished_at,
            duration=pipeline.duration,
            jobs=job_runs,
        )

    async def delete_project(self, project_id: int, wait: bool = True):
        await force_async(self.gitlab.projects.delete)(project_id)
        if wait:
            try:
                proj = await self._get_project(project_id)
                while proj is not None:
                    LOG.info(f"Waiting for project {project_id} to delete")
                    time.sleep(2)
                    proj = await self._get_project(project_id)
            except ValueError:
                LOG.info(f"Project {project_id} successfully deleted")
                return

    async def delete_project_by_name(self, project_name: str, wait: bool = True):
        settings = get_settings()
        projects = await force_async(self.gitlab.projects.list)(
            search=project_name, owned=True
        )
        LOG.info(f"Found projects: {projects}")
        for proj in projects:
            group_id = proj.namespace["id"]
            LOG.info(f"Checking project: {proj.id}, {proj.name}, {group_id}")
            if proj.name == project_name and group_id == settings.GITLAB_GROUP_ID:
                LOG.info(f"Deleting project: {proj.id}")
                await self.delete_project(proj.id, wait)
                return
        raise ValueError(f"No project found with the name {project_name}")



from abc import ABC
from typing import Iterable, List, Optional, Tuple

from fa_common import CamelModel, DatabaseError
from fa_common.db import DocumentDBModel, WriteResult


class AbstractMeta(ABC):
    exclude: set = set()
    model: Optional[DocumentDBModel] = None
    write_only_fields: set = set()
    read_only_fields: set = set()


class Serializer(CamelModel):
    """
    Base Serializer class.
    Almost ALWAYS should be used in conjunction with
    `fa_common.core.serializers.openapi.patch` decorator to correctly handle
    inherited model fields and OpenAPI Schema generation with `response_model`.
    Responsible for sanitizing data & converting JSON to & from DocumentDBModel.
    Contains supplemental function, related to DocumentDBModel,
    mostly proxied to corresponding functions inside model (ex. save, update)
    Heavily uses `Meta` class for fine-tuning input & output. Main fields are:
        * exclude - set of fields that are excluded when serializing to dict
                    and sanitizing list of dicts
        * model - class of the DocumentDBModel to use, inherits fields from it
        * write_only_fields - set of fields that can be accepted in request,
                              but excluded when serializing to dict
        * read_only_fields - set of fields that cannot be accepted in request,
                              but included when serializing to dict
    Example usage:
    .. code-block:: python
        app = FastAPI()
        class SomeModel(DocumentDBModel):
            field1: str
        @openapi.patch
        class SomeSerializer(Serializer):
            read_only1: str = "const"
            write_only2: int
            not_visible: str = "42"
            class Meta:
                model = SomeModel
                exclude = {"not_visible"}
                write_only_fields = {"write_only2"}
                read_only_fields = {"read_only1"}
        @app.get("/", response_model=SomeSerializer.response_model)
        async def root(serializer: SomeSerializer):
            model_instance = await serializer.save()
            return model_instance.dict()
    POST-ing to this route following JSON:
    .. code-block:: json
        {"read_only1": "a", "write_only2": 123, "field1": "b"}
    Should return following response:
    .. code-block:: json
        {"id": 1, "field1": "b", "read_only1": "const"}
    """

    @classmethod
    def sanitize_list(cls, iterable: Iterable) -> List[dict]:
        """
        Sanitize list of rows that comes from DB to not include `exclude` set.
        :param iterable: sequence of dicts with model fields (from rows in DB)
        :return: list of cleaned, without `excluded`, dicts with model rows
        """

        def clean_d(d):
            if hasattr(cls.Meta, "exclude"):
                for e in cls.Meta.exclude:
                    d.pop(e, None)
                return d
            return d

        return list(map(lambda x: clean_d(x), iterable))

    async def save(
        self, include: set = None, exclude: set = None, rewrite_fields: dict = None,
    ) -> Optional[str]:
        """
        If we have `model` attribute in Meta, it populates model with data
        and saves it in DB, returning instance of model.
        :param rewrite_fields: dict of fields with values that override any
                other values for these fields right before inserting into DB.
                This is useful when you need to set some value explicitly
                based on request (e.g. user or token).
        :param include: fields to include from model in DB insert command
        :param exclude: fields to exclude from model in DB insert command
        :return: id (str) that was saved
        """
        if hasattr(self, "Meta") and getattr(self.Meta, "model", None) is not None:
            instance = self.Meta.model.__class__(**self.__values__)
            ret = await instance.save(
                include=include, exclude=exclude, rewrite_fields=rewrite_fields
            )
            return ret
        return None

    async def update_one(
        self, _id: str, _data: dict, skip_defaults: bool = True
    ) -> WriteResult:
        """
        If we have `model` attribute in Meta, it proxies filters & update data
        and after that returns actual result of update operation.
        :return: result of update operation
        """
        if hasattr(self, "Meta") and getattr(self.Meta, "model", None) is not None:
            fields = self.dict(skip_defaults=skip_defaults)
            if "exclude" in fields:
                for e in fields["exclude"]:
                    if e in _data:
                        del _data[e]

            return await self.Meta.model.update_one(_id, _data)  # type: ignore
        raise DatabaseError(f"No Meta.model is defined for {self.__class__}")

    async def update_many(
        self, data: List[Tuple[str, dict]], skip_defaults: bool = True
    ) -> WriteResult:
        """
        If we have `model` attribute in Meta, it proxies filters & update data
        and after that returns actual result of update operation.
        :return: result of update many operation
        """
        if hasattr(self, "Meta") and getattr(self.Meta, "model", None) is not None:

            fields = self.dict(skip_defaults=skip_defaults)
            for k, v in data:
                if "exclude" in fields:
                    for e in fields["exclude"]:
                        if e in v:
                            del v[e]

            return await self.Meta.model.update_many(data=data)  # type: ignore
        raise DatabaseError(f"No Meta.model is defined for {self.__class__}")

    def dict(self, *args, **kwargs) -> dict:
        """
        Removes excluded fields based on `Meta` and `kwargs`
        :return: dict of serializer data fields
        """
        exclude = kwargs.get("exclude")
        if not exclude:
            exclude = set()

        if hasattr(self.Meta, "exclude") and self.Meta.exclude:
            exclude.update(self.Meta.exclude)

        if hasattr(self.Meta, "write_only_fields") and self.Meta.write_only_fields:
            exclude.update(self.Meta.write_only_fields)

        kwargs.update({"exclude": exclude})
        original = super().dict(*args, **kwargs)
        return original

    class Meta(AbstractMeta):
        ...


# class CamelSerializer(Serializer, CamelModel):
#     ...


class ModelSerializer(Serializer):
    """
    Left as a proxy for correct naming until we figure out how to inherit
    all the specific to model-handling methods and fields directly in here.
    """

    ...



import os
import requests
from io import BytesIO

import pytest
from fastapi import FastAPI, UploadFile

from pathlib import Path

from fa_common import create_app, start_app, utils
from fa_common.storage import get_storage_client

from .conftest import clean_env, TEST_USER_BUCKET

dirname = os.path.dirname(__file__)
test_data_path = os.path.join(dirname, "data")


async def setup_storage(path: str):

    app = create_app(env_path=path)
    await start_app(app)
    utils.current_app = app

    assert isinstance(app, FastAPI)
    client = get_storage_client()
    await client.make_bucket(TEST_USER_BUCKET)
    assert await client.bucket_exists(TEST_USER_BUCKET)
    return client


async def exercise_storage(path: str):

    client = await setup_storage(path)

    filename = "file1.txt"
    up_file = UploadFile(
        filename=filename, file=open(os.path.join(test_data_path, filename), "rb")
    )
    # Test file upload
    file_ref = await client.upload_file(up_file, TEST_USER_BUCKET, "files")
    assert file_ref is not None

    # Test list files
    files = await client.list_files(TEST_USER_BUCKET, "files")
    assert files is not None and len(files) > 0

    # Test get file returns file correctly
    file_stream = await client.get_file(TEST_USER_BUCKET, "files/file1.txt")

    with open(os.path.join(test_data_path, filename), "rb") as fin:
        test_file_bytes = BytesIO(fin.read())
        assert (
            file_stream is not None
            and file_stream.getvalue() == test_file_bytes.getvalue()
        )

    # Test file exists Negative
    file_stream2 = await client.file_exists(TEST_USER_BUCKET, "files/file3.txt")
    assert file_stream2 is False

    # Test file exists Positive
    file_stream2 = await client.file_exists(TEST_USER_BUCKET, "files/file1.txt")
    assert file_stream2 is True

    try:
        download_ref = await client.create_temp_file_url(
            TEST_USER_BUCKET, "files/file1.txt"
        )

        r = requests.get(download_ref.url)  # type: ignore
        target_path = Path(os.path.join(dirname, "output", download_ref.name))
        target_path.parent.mkdir(parents=True, exist_ok=True)
        with open(target_path, "wb") as _file:
            _file.write(r.content)
        assert os.path.exists(os.path.join(dirname, "output", download_ref.name))
    except NotImplementedError:
        pass

    # Test Copy a file
    await client.copy_file(
        TEST_USER_BUCKET, "files/file1.txt", TEST_USER_BUCKET, "files/file1_copy.txt"
    )
    file1_exists = await client.file_exists(TEST_USER_BUCKET, "files/file1.txt")
    file1_copy_exists = await client.file_exists(
        TEST_USER_BUCKET, "files/file1_copy.txt"
    )
    assert file1_exists and file1_copy_exists

    # Test Rename a file
    await client.rename_file(
        TEST_USER_BUCKET, "files/file1.txt", "files/file1_renamed.txt"
    )
    file1_exists = await client.file_exists(TEST_USER_BUCKET, "files/file1.txt")
    file1_r_exists = await client.file_exists(
        TEST_USER_BUCKET, "files/file1_renamed.txt"
    )
    assert file1_exists is False and file1_r_exists

    # Test get file returns None when file doesn't exist
    file_stream3 = await client.get_file(TEST_USER_BUCKET, "files/file3.txt")
    assert file_stream3 is None

    filename = "file2.txt"
    up_file2 = UploadFile(
        filename=filename, file=open(os.path.join(test_data_path, filename), "rb")
    )
    file_ref = await client.upload_file(up_file2, TEST_USER_BUCKET, "files/folder")

    assert file_ref is not None

    file_ref = await client.upload_string(
        "test_string weee", TEST_USER_BUCKET, "files/test_string.txt"
    )
    assert file_ref is not None
    assert await client.file_exists(
        TEST_USER_BUCKET, "files/test_string.txt"
    ), "test_string.txt Is missing from bucket"

    # Test folder exists
    assert await client.folder_exists(
        TEST_USER_BUCKET, "files/"
    ), "Folder Is missing from bucket or testing for folders is not working"

    # Test non existing folder does not exists
    assert not await client.folder_exists(
        TEST_USER_BUCKET, "files_not/"
    ), "Testing for non existing folder failed"

    # Test deleting a folder
    assert await client.file_exists(
        TEST_USER_BUCKET, "files/folder/file2.txt"
    ), "File2 Is missing from bucket"
    await client.delete_file(TEST_USER_BUCKET, "files/folder", True)
    assert not await client.file_exists(
        TEST_USER_BUCKET, "files/folder/file2.txt"
    ), "File2 Is in bucket when it should be deleted"
    assert not await client.file_exists(
        TEST_USER_BUCKET, "files/folder/"
    ), "Folder still exists when it should be deleted"

    await client.delete_bucket(TEST_USER_BUCKET)
    assert not await client.bucket_exists(TEST_USER_BUCKET)

    clean_env()


@pytest.mark.asyncio
async def test_gcp_file_ref(env_path):
    client = await setup_storage(env_path)
    file_ref = await client.get_file_ref(TEST_USER_BUCKET, "files/folder/temp_file.csv")
    assert file_ref.gs_uri is not None
    assert not await client.file_exists(TEST_USER_BUCKET, "files/folder/temp_file.csv")

    clean_env()


@pytest.mark.asyncio
async def test_firbase_storage(env_path):
    await exercise_storage(env_path)


@pytest.mark.asyncio
async def test_minio_storage(minio_env_path):
    await exercise_storage(minio_env_path)



import os

import pytest
import socket

from fa_common.auth import AuthUser
from fa_common.workflow import get_workflow_client
from fa_common import force_sync, logger as LOG
from dotenv import load_dotenv

dirname = os.path.dirname(__file__)
env_file = os.path.join(dirname, ".env")
secrets_env_file = os.path.join(dirname, ".env.secret")
minio_env_file = os.path.join(dirname, ".env-minio")

hostname = socket.gethostname().replace("_", "-").replace(".", "-").lower()
TEST_GITLAB_PROJECT = f"test-user-{hostname}"
TEST_GITLAB_BRANCH = "test_project"
TEST_USER_BUCKET = f"test-user-{hostname}"


def get_env_file() -> str:
    clean_env()
    if os.path.exists(secrets_env_file):
        load_dotenv(dotenv_path=env_file)
        return secrets_env_file
    else:
        return env_file


@pytest.fixture(scope="module")
def env_path(scope="module") -> str:
    return get_env_file()


@pytest.fixture(scope="module")
def minio_env_path(scope="module") -> str:
    clean_env()
    if os.path.exists(secrets_env_file):
        load_dotenv(dotenv_path=minio_env_file)
        return secrets_env_file
    else:
        return minio_env_file


@pytest.fixture
def ensure_no_test_gitlab():
    client = get_workflow_client()
    try:
        project = force_sync(client._get_project_by_name)(TEST_GITLAB_PROJECT)
    except ValueError:
        LOG.info(f"No project called {TEST_GITLAB_PROJECT}")
        return

    force_sync(client.delete_project)(project.id, True)
    LOG.info(f"Deleted {project.id}")


@pytest.fixture
def ensure_test_gitlab():
    client = get_workflow_client()
    try:
        return (force_sync(client._get_project_by_name)(TEST_GITLAB_PROJECT)).id
    except Exception as err:
        LOG.info(err)
        project_id = force_sync(client.create_project)(TEST_GITLAB_PROJECT)
        force_sync(client.create_branch)(project_id, TEST_GITLAB_BRANCH)
        return project_id


def clean_env():
    os.environ.pop("STORAGE_TYPE", None)
    os.environ.pop("BUCKET_PREFIX", None)
    os.environ.pop("BUCKET_NAME", None)
    os.environ.pop("BUCKET_USER_FOLDER", None)
    os.environ.pop("DATABASE_TYPE", None)
    os.environ.pop("GOOGLE_APPLICATION_CREDENTIALS", None)
    os.environ.pop("ENABLE_FIREBASE", None)
    os.environ.pop("MINIO_ENDPOINT", None)
    os.environ.pop("MINIO_ACCESS_KEY", None)
    os.environ.pop("MINIO_SECRET_KEY", None)


def pytest_runtest_makereport(item, call):
    if "incremental" in item.keywords:
        if call.excinfo is not None:
            parent = item.parent
            parent._previousfailed = item


def pytest_runtest_setup(item):
    if "incremental" in item.keywords:
        previousfailed = getattr(item.parent, "_previousfailed", None)
        if previousfailed is not None:
            pytest.xfail("previous test failed ({})".format(previousfailed.name))


async def get_test_user() -> AuthUser:
    payload = {
        "sub": f"{hostname}-sub",
        "name": "Test User",
        "email": "test.user@test.com",
        "nickname": "test.user",
        "email_verified": True,
        "picture": None,
        "updated_at": None,
        "scopes": [],
    }

    return AuthUser(**payload)



import functools
from typing import Awaitable, Callable, Type

from fastapi import FastAPI
from fastapi.middleware.gzip import GZipMiddleware
from starlette.middleware.cors import CORSMiddleware

from .config import DatabaseType, Settings, StorageType, get_settings
from .exception_handlers import setup_exception_handlers
from .exceptions import (
    BadRequestError,
    DatabaseError,
    ForbiddenError,
    HTTPException,
    NotFoundError,
    NotImplementedError,
    StorageError,
    UnauthorizedError,
    UnknownError,
)
from .models import CamelModel, FileDownloadRef
from .responses import UJSONResponse
from .utils import (
    async_get,
    force_async,
    force_sync,
    get_current_app,
    get_logger,
    get_now,
    get_remote_schema,
    get_timezone,
    logger,
    resolve_dotted_path,
    sizeof_fmt,
)


def create_app(
    env_path: str = None,
    disable_gzip: bool = False,
    on_start: Callable[[FastAPI], Awaitable[None]] = None,
    on_stop: Callable[[FastAPI], Awaitable[None]] = None,
) -> FastAPI:
    settings = get_settings(env_path)

    if settings.SENTRY_DSN is not None:
        import sentry_sdk

        sentry_sdk.init(settings.SENTRY_DSN)

    app = FastAPI(
        title=settings.PROJECT_NAME,
        openapi_url=f"{settings.API_PRE_PATH}/openapi.json",
    )

    # CORS
    origins = []

    # Set all CORS enabled origins
    if settings.BACKEND_CORS_ORIGINS:
        origins_raw = settings.BACKEND_CORS_ORIGINS
        for origin in origins_raw:
            use_origin = origin.strip()
            origins.append(use_origin)
        logger.info(f"Allowing Origins {origins}")
        app.add_middleware(
            CORSMiddleware,
            allow_origins=origins,
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

    # Adds support for GZIP response
    if not disable_gzip:
        app.add_middleware(GZipMiddleware, minimum_size=5000)

    setup_exception_handlers(app)

    @app.on_event("startup")
    async def on_start_app() -> None:
        await start_app(app)
        if on_start is not None:
            await on_start(app)

    @app.on_event("shutdown")
    @logger.catch
    async def stop_app() -> None:
        logger.info("Stopping App")
        # await close_db_connection(app)
        if on_stop is not None:
            await on_stop(app)

    return app


async def start_app(app: FastAPI):
    from fa_common.storage import setup_storage
    from fa_common.db import setup_db, create_indexes

    if get_settings().USE_FIREBASE:
        import firebase_admin

        if not len(firebase_admin._apps):
            firebase_admin.initialize_app()

    setup_db(app)
    setup_storage(app)
    try:
        from fa_common.workflow import setup_gitlab

        setup_gitlab(app)
    except ValueError:
        logger.info(
            "Gitlab dependencies are missing, if you are planning to use workflows make sure the optional "
            + "dependencies are installed"
        )
        pass  # App probably doesn't want to use gitlab

    await create_indexes()



import abc
from typing import Any, List, Optional, Tuple, Type

from .models import (
    DBIndex,
    DeleteResult,
    DocumentDBModel,
    FireOffset,
    SortOrder,
    WhereCondition,
    WriteResult,
)


class BaseClient(abc.ABC):
    @abc.abstractmethod
    def generate_id(self, collection_name: str) -> str:
        pass

    @abc.abstractmethod
    async def insert(
        self, model: DocumentDBModel, include=None, exclude=None, session: Any = None,
    ) -> WriteResult:
        pass

    @abc.abstractmethod
    async def count(self, model: Type[DocumentDBModel], session: Any = None) -> int:
        pass

    @abc.abstractmethod
    async def get(
        self, model: Type[DocumentDBModel], _id: str, session: Any = None
    ) -> Optional[DocumentDBModel]:
        pass

    @abc.abstractmethod
    async def list(
        self,
        model: Type[DocumentDBModel],
        where: List[WhereCondition] = [],
        _limit: int = 0,
        _sort: List[Tuple[str, SortOrder]] = None,
        mongo_offset: int = 0,
        fire_offset: FireOffset = None,
    ) -> List[DocumentDBModel]:
        """
        List documents in a collection based on specified conditions, note this interface is limiting
        query to the smaller set of features (currently Firestore).

        Arguments:
            model {DocumentDBModel} -- [description]

        Keyword Arguments:
            where {List[WhereCondition]} -- [description] (default: {[]})
            session {Any} -- [description] (default: {None})
            _offset {int} -- [description] (default: {0})
            _limit {int} -- [description] (default: {0})
            _sort {List[Tuple} -- [description] (default: {None})

        Returns:
            Cursor -- [description]
        """
        pass

    @abc.abstractmethod
    async def delete(
        self, model: Type[DocumentDBModel], _id: str, session: Any = None
    ) -> DeleteResult:
        pass

    @abc.abstractmethod
    async def update_one(
        self, model: Type[DocumentDBModel], id: str, data: dict, session: Any = None,
    ) -> WriteResult:
        pass

    @abc.abstractmethod
    async def create_indexes(self, index=List[DBIndex]) -> Optional[List[str]]:
        pass

    # FIXME: Need to add this back once it has been fixed
    # @abc.abstractmethod
    # async def update_many(
    #     self, model: Type[DocumentDBModel], data: List[Tuple[str, dict]]
    # ) -> WriteResult:
    #     pass



from typing import Any, List, Optional, Type

from bson import CodecOptions
from motor.motor_asyncio import AsyncIOMotorClient

# from pymongo.client_session import ClientSession
from pymongo.collection import Collection

# from pymongo.cursor import Cursor
# from pymongo.results import InsertOneResult, UpdateResult

from fa_common import get_current_app, get_timezone
from fa_common import logger as LOG

from .base_client import BaseClient
from .models import (
    DeleteResult,
    FireOffset,
    Operator,
    SortOrder,
    WhereCondition,
    WriteResult,
    DocumentDBModel,
)


class MongoDBClient(BaseClient):
    """
    Singleton client for interacting with MongoDB.
    Operates mostly using models, specified when making DB queries.

    Implements only part of internal `motor` methods, but can be populated more.

    Please don't use it directly, use `scidra.core.db.utils.get_db_client`.
    """

    __instance = None
    mongodb: AsyncIOMotorClient = None

    def __new__(cls) -> "MongoDBClient":
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
            app = get_current_app()
            tzinfo = get_timezone()
            cls.__instance.codec_options = CodecOptions(tz_aware=True, tzinfo=tzinfo)
            cls.__instance.mongodb = app.mongodb  # type: ignore
        return cls.__instance

    def get_collection(self, collection_name: str) -> Collection:
        return self.mongodb.get_collection(
            collection_name, codec_options=self.codec_options
        )

    async def insert(
        self, model: DocumentDBModel, include=None, exclude=None, session: Any = None,
    ) -> WriteResult:
        data = model.dict(include=include, exclude=exclude)
        data["_id"] = data.pop("id")
        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        result: InsertOneResult = await collection.insert_one(data, session=session)

        LOG.info("Insert Result: {}", result)
        # FIXME: Correct result return
        return WriteResult()

    async def count(self, model: Type[DocumentDBModel], session: Any = None) -> int:

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        res = await collection.count_documents(session=session)
        return res

    async def delete(
        self, model: Type[DocumentDBModel], _id: str, session: Any = None
    ) -> DeleteResult:

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        res = await collection.delete_one(_id=_id, session=session)
        return DeleteResult()

    async def update_one(
        self, model: Type[DocumentDBModel], _id: str, data: dict, session: Any = None
    ) -> WriteResult:
        data.pop("id", None)
        if _id is not None:
            data["_id"] = _id

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        await collection.update_one(filter={"_id": _id}, update=data, session=session)
        return WriteResult()

    # FIXME: Add this back once the firstore version has been fixed and added
    # async def update_many(
    #     self, model: Type[DocumentDBModel], data: List[Tuple[str, dict]]
    # ) -> WriteResult:
    #     _id = filter_kwargs.pop("id", None)
    #     if _id is not None:
    #         filter_kwargs["_id"] = _id

    #     collection_name = model.get_db_collection()
    #     collection = self.get_collection(collection_name)
    #     res = await collection.update_many(filter_kwargs, kwargs, session=session)
    #     return res

    async def get(
        self, model: Type[DocumentDBModel], _id: str, session: Any = None
    ) -> Optional[DocumentDBModel]:

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        res = await collection.find_one(_id=_id, session=session)
        if res is None:
            return None

        return model(**res)

    async def list(
        self,
        model: Type[DocumentDBModel],
        where: List[WhereCondition] = [],
        _limit: int = 0,
        _sort: List[Tuple[str, SortOrder]] = None,
        mongo_offset: int = 0,
        fire_offset: FireOffset = None,
    ) -> List[DocumentDBModel]:

        _id = kwargs.pop("id", None)
        if _id is not None:
            kwargs["_id"] = _id

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)

        filter_dict = {}
        for wc in where:
            if wc.operator is Operator.EQUALS:
                filter_dict[wc.field] = wc.value
            elif wc.operator is Operator.LT:
                filter_dict[wc.field] = {"$lt": wc.value}
            elif wc.operator is Operator.GT:
                filter_dict[wc.field] = {"$gt": wc.value}
            # TODO: Other array filters

        finder = collection.find(filter_dict, skip=mongo_offset, limit=_limit)
        if _sort is not None:
            for sort in _sort:
                finder.sort(sort[0], sort[1])
        results = await finder.to_list()
        models: List[DocumentDBModel] = []
        for res in results:
            models.append(model(**res))
        return models



from .models import (
    DBIndex,
    DeleteResult,
    DocumentDBModel,
    DocumentDBTimeStampedModel,
    FireOffset,
    Operator,
    SortOrder,
    WhereCondition,
    WriteResult,
)
from .utils import create_indexes, get_db_client, setup_db



import abc
from io import BytesIO
from typing import List, Optional, Union

from fastapi import UploadFile
from fa_common import FileDownloadRef

from .model import File


class BaseClient(abc.ABC):
    @abc.abstractmethod
    async def make_bucket(self, name: str) -> None:
        pass

    @abc.abstractmethod
    async def bucket_exists(self, name: str) -> bool:
        pass

    @abc.abstractmethod
    async def delete_bucket(self, name: str):
        pass

    @abc.abstractmethod
    async def list_files(self, bucket_name: str, parent_path: str = "") -> List[File]:
        pass

    @abc.abstractmethod
    async def upload_file(
        self, file: UploadFile, bucket_name: str, parent_path: str = "",
    ) -> File:
        pass

    @abc.abstractmethod
    async def upload_string(
        self,
        string: Union[str, bytes],
        bucket_name: str,
        file_path: str,
        content_type="text/plain",
    ) -> File:
        pass

    @abc.abstractmethod
    async def get_file_ref(self, bucket_name: str, file_path: str) -> Optional[File]:
        pass

    @abc.abstractmethod
    async def get_file(self, bucket_name: str, file_path: str) -> Optional[BytesIO]:
        pass

    @abc.abstractmethod
    async def file_exists(self, bucket_name: str, file_path: str) -> bool:
        pass

    @abc.abstractmethod
    async def folder_exists(self, bucket_name: str, path: str) -> bool:
        pass

    @abc.abstractmethod
    async def delete_file(
        self, bucket_name: str, file_path: str, recursive: bool = False
    ) -> None:
        """
        Deletes a file or folder from the specified bucket

        Arguments:
            bucket_name {str} -- [description]
            file_path {str} -- [description]

        Keyword Arguments:
            recursive {bool} -- Deletes all child & folders files from a non empty folder (default: {False})

        """
        pass

    @abc.abstractmethod
    async def rename_file(
        self, bucket_name: str, file_path: str, new_file_path: str
    ) -> None:
        pass

    @abc.abstractmethod
    async def copy_file(
        self, from_bucket: str, from_path: str, to_bucket: str, to_path: str
    ) -> None:
        pass

    @abc.abstractmethod
    async def create_temp_file_url(
        self, bucket: str, path: str, expire_time_hours: int = 3
    ) -> File:
        """Enable file to be downloaded without auth via a URL temporarily
        """
        pass



from fastapi import FastAPI

from fa_common import StorageType, get_settings, logger

from .base_client import BaseClient

# from minio.error import ResponseError


def setup_storage(app: FastAPI) -> None:
    settings = get_settings()
    if settings.STORAGE_TYPE == StorageType.MINIO:
        from minio import Minio

        if (
            settings.MINIO_SECRET_KEY is None
            or settings.MINIO_ACCESS_KEY is None
            or settings.MINIO_ENDPOINT is None
        ):
            raise ValueError("Missing minio settings from env variables")

        minioClient = Minio(
            settings.MINIO_ENDPOINT,
            access_key=settings.MINIO_ACCESS_KEY,
            secret_key=settings.MINIO_SECRET_KEY.get_secret_value(),
            secure=settings.MINIO_SSL,
        )
        app.minio = minioClient  # type: ignore
    elif (
        settings.STORAGE_TYPE == StorageType.GCP_STORAGE
        or settings.STORAGE_TYPE == StorageType.FIREBASE_STORAGE
    ):
        from google.cloud import storage

        # Uses GOOGLE_APPLICATION_CREDENTIALS Env Var
        gcp_storage_client = storage.Client()
        app.gcp_storage = gcp_storage_client  # type: ignore
    elif settings == StorageType.NONE:
        logger.info("Storage set to NONE and cannot be used")
        return
    else:
        raise ValueError("STORAGE_TYPE Setting is not a valid storage option.")


def get_storage_client() -> BaseClient:
    if get_settings().STORAGE_TYPE == StorageType.MINIO:
        from .minio_client import MinioClient

        return MinioClient()
    elif get_settings().STORAGE_TYPE == StorageType.GCP_STORAGE:
        from .gcp_client import GoogleStorageClient

        return GoogleStorageClient()
    elif get_settings().STORAGE_TYPE == StorageType.FIREBASE_STORAGE:
        from .gcp_client import FirebaseStorageClient

        return FirebaseStorageClient()

    raise ValueError("STORAGE_TYPE Setting is not a valid storage option.")



import os
import pytest

from fa_common import create_app, start_app, utils, force_sync
from fa_common.workflow import (
    run_job,
    ScidraModule,
    get_job_run,
    get_job_log,
    create_workflow_project,
    WorkflowProject,
    get_workflow_client,
)
from fa_common.storage import File
from .conftest import get_env_file, TEST_GITLAB_PROJECT, TEST_GITLAB_BRANCH

dirname = os.path.dirname(__file__)
test_data_path = os.path.join(dirname, "data")

app = create_app(env_path=get_env_file())
force_sync(start_app)(app)
utils.current_app = app


@pytest.mark.asyncio
async def test_gitlab_create(ensure_no_test_gitlab):
    workflow_project = await create_workflow_project(
        TEST_GITLAB_PROJECT, TEST_GITLAB_BRANCH
    )

    assert workflow_project.gitlab_project_id > 0
    assert workflow_project.name == TEST_GITLAB_BRANCH


@pytest.mark.asyncio
async def test_job_create(ensure_test_gitlab):
    project_id = ensure_test_gitlab
    assert project_id > 0

    project = WorkflowProject(
        name=TEST_GITLAB_BRANCH,
        user_id=TEST_GITLAB_PROJECT,
        gitlab_project_id=project_id,
    )

    module = ScidraModule(
        name="noise-detection",
        docker_image="registry.gitlab.com/csiro-geoanalytics/gpt/modules/noise-detection:latest",
    )
    job_data = {
        "window_channels": 4,
        "window_size": 20,
        "channel_step": 2,
        "window_step": 10,
        "variation_threshold": 0.01,
        "crossover_threshold": 0.003,
    }
    files = [
        File(
            id="data_file",
            name="data_file.csv",
            url="https://firebasestorage.googleapis.com/v0/b/gp-toolkit-staging.appspot.com/o/"
            + "unit_test_data%2Fnd_test_data.csv?alt=media&token=96a65aa1-9e34-4a37-9bc7-7eac4f14847e",
        )
    ]

    job_run = await run_job(
        project,
        "Sync Unit Test",
        module,
        job_data,
        runner="csiro-swarm",
        files=files,
        sync=True,
    )

    assert job_run.status == "success"
    assert job_run.output is not None

    job_run2 = await get_job_run(project_id, job_run.id, False)

    assert job_run2.id == job_run.id

    job_logs = await get_job_log(project_id, job_run.id)
    assert job_logs is not None


@pytest.mark.asyncio
async def test_job_create_async(ensure_test_gitlab):
    project_id = ensure_test_gitlab
    assert project_id > 0

    project = WorkflowProject(
        name=TEST_GITLAB_BRANCH,
        user_id=TEST_GITLAB_PROJECT,
        gitlab_project_id=project_id,
    )

    module = ScidraModule(
        name="noise-detection",
        docker_image="registry.gitlab.com/csiro-geoanalytics/gpt/modules/noise-detection:latest",
    )
    job_data = {
        "window_channels": 4,
        "window_size": 20,
        "channel_step": 2,
        "window_step": 10,
        "variation_threshold": 0.01,
        "crossover_threshold": 0.003,
    }
    files = [
        File(
            id="data_file",
            name="data_file.csv",
            url="https://firebasestorage.googleapis.com/v0/b/gp-toolkit-staging.appspot.com/o/"
            + "unit_test_data%2Fnd_test_data.csv?alt=media&token=96a65aa1-9e34-4a37-9bc7-7eac4f14847e",
        )
    ]

    job_run = await run_job(
        project, "Async Unit Test", module, job_data, runner="csiro-swarm", files=files
    )

    job_run2 = await get_job_run(project_id, job_run.id)

    assert job_run2.id == job_run.id


# @pytest.mark.asyncio
# async def test_large_job_create_async(ensure_test_gitlab):
#     project_id = ensure_test_gitlab
#     assert project_id > 0

#     project = WorkflowProject(
#         name=TEST_GITLAB_BRANCH,
#         user_id=TEST_GITLAB_PROJECT,
#         gitlab_project_id=project_id,
#         timezone="Australia/Perth",
#     )

#     module = ScidraModule(
#         name="galei-inversion",
#         docker_image="registry.gitlab.com/csiro-geoanalytics/gpt/modules/galei-inversion-module:latest",
#         cpu_limit=4,
#         cpu_request=2,
#         memory_limit_gb=16,
#         memory_request_gb=4,
#     )
#     job_data = {"test": "subset"}

#     job_run = await run_job(
#         project, "Inversion Test", module, job_data, runner="csiro-easi-hub"
#     )

#     assert job_run.id is not None


@pytest.mark.asyncio
async def test_delete_workflow_project():
    client = get_workflow_client()
    try:
        project = await client._get_project_by_name(TEST_GITLAB_PROJECT)
    except ValueError:
        return

    await client.delete_project(project.id, False)



from typing import Type

from .utils import FieldGenerationMode, gen_model


def patch(cls: Type) -> Type:
    """
    Decorator for `Serializer` classes to handle inheritance from models,
    read- and write-only fields, combining `Meta`s.
    For more info see `gen_model` method.
    :param cls: serializer class (model or regular)
    :return: wrapped class, which is newly generated pydantic's `BaseModel`
    """
    return gen_model(cls, mode=FieldGenerationMode.REQUEST)



import functools
import importlib
import logging
import os
import sys
from datetime import datetime
from typing import Any, Optional

import aiohttp
import pytz
from fastapi import FastAPI
from loguru import logger as loguru_logger

from .config import get_settings

# from scidra.main import app


current_app: Optional[FastAPI] = None


def force_async(fn):
    """
    turns a sync function to async function using threads
    """
    from concurrent.futures import ThreadPoolExecutor
    import asyncio

    pool = ThreadPoolExecutor()

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        future = pool.submit(fn, *args, **kwargs)
        return asyncio.wrap_future(future)  # make it awaitable

    return wrapper


def force_sync(fn):
    """
    turn an async function to sync function
    """
    import asyncio

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        res = fn(*args, **kwargs)
        if asyncio.iscoroutine(res):
            return asyncio.get_event_loop().run_until_complete(res)
        return res

    return wrapper


def sizeof_fmt(num: Optional[Any], suffix="B") -> str:
    if num is None:
        return ""
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, "Y", suffix)


def resolve_dotted_path(path: str) -> Any:
    """
    Retrieves attribute (var, function, class, etc.) from module by dotted path.
    .. code-block:: python
        from datetime.datetime import utcnow as default_utcnow
        utcnow = resolve_dotted_path('datetime.datetime.utcnow')
        assert utcnow == default_utcnow
    :param path: dotted path to the attribute in module
    :return: desired attribute or None
    """
    splitted = path.split(".")
    if len(splitted) <= 1:
        return importlib.import_module(path)
    module, attr = ".".join(splitted[:-1]), splitted[-1]
    module = importlib.import_module(module)  # type: ignore
    return getattr(module, attr)


class PropagateHandler(logging.Handler):
    def emit(self, record):
        logging.getLogger("gunicorn.error").handle(record)


def get_logger() -> Any:
    """
    Gets logger that will be used throughout this whole library.
    First it finds and imports the logger, then if it can be configured
    using loguru-compatible config, it does so.

    :return: desired logger (pre-configured if loguru)
    """
    lib_logger = loguru_logger

    # Check whether it is loguru-compatible logger
    if hasattr(lib_logger, "configure"):
        lib_logger.remove()
        is_gunicorn = "gunicorn" in os.environ.get("SERVER_SOFTWARE", "")
        is_gae = os.environ.get("GAE_APPLICATION", None) is not None

        sink = PropagateHandler() if is_gunicorn else sys.stdout
        logger_config = {
            "handlers": [
                {
                    "sink": sink,
                    "level": get_settings().LOGGING_LEVEL,
                    "format": "<level>{level: <8}</level> | "
                    "<cyan>{name}</cyan>:<cyan>{function}</cyan>:"
                    "<cyan>{line}</cyan> -"
                    " <level>{message}</level>",
                }
            ]
        }
        if is_gae:
            from google.cloud import logging as g_logging  # type: ignore

            client = g_logging.Client()
            client.setup_logging()

        lib_logger.configure(**logger_config)  # type: ignore

    return lib_logger


logger = get_logger()


async def async_get(url, auth: str = None):
    async with aiohttp.ClientSession() as session:
        headers = {}
        if auth is not None:
            headers = {"Authorization": auth}
        async with session.get(url, headers=headers) as response:
            return await response.json()


def get_current_app() -> FastAPI:
    """
    Retrieves FastAPI app instance from the path, specified in project's conf.
    :return: FastAPI app
    """
    global current_app
    if current_app is None:
        logger.info("Retrieving app from dotted path")
        current_app = resolve_dotted_path(get_settings().FASTAPI_APP)
    return current_app


async def get_remote_schema(host) -> str:
    """
    Retrieves the open api json for the given url
    :return: json schema dict
    """
    async with aiohttp.ClientSession() as session:
        async with session.get(host + "/openapi.json") as r:
            return await r.json()


def get_now() -> datetime:
    """
    Retrieves `now` function from the path, specified in project's conf.
    :return: datetime of "now"
    """
    return datetime.now(tz=get_timezone())


def get_timezone():
    """
    Retrieves timezone name from settings and tries to create tzinfo from it.
    :return: tzinfo object
    """
    return pytz.timezone(get_settings().TZ)



from enum import Enum
from typing import Type, List, Set, Mapping, Tuple, Sequence

from pydantic import Required, create_model
from pydantic.fields import (
    SHAPE_LIST,
    SHAPE_SET,
    SHAPE_MAPPING,
    SHAPE_TUPLE,
    SHAPE_TUPLE_ELLIPSIS,
    SHAPE_SEQUENCE,
)

from .common import AbstractMeta, Serializer


class FieldGenerationMode(int, Enum):
    """
    Defines modes in which fields of decorated serializer should be generated.
    """

    REQUEST = 1
    RESPONSE = 2


def gen_model(cls: Type, mode: FieldGenerationMode):
    """
    Generate `pydantic.BaseModel` based on fields in Serializer class,
    its Meta class and possible Model class.
    :param cls: serializer class (could be modelserializer or regular one)
    :param mode: field generation mode
    :return: newly generated `BaseModel` from fields in Model & Serializer
    """
    _fields = {}

    _Meta = getattr(cls, "Meta", type("Meta"))
    Meta = type("Meta", (_Meta, AbstractMeta), {})

    Config = getattr(cls, "Config", getattr(Serializer, "Config"))

    if mode == FieldGenerationMode.RESPONSE:
        excluded = Meta.exclude | Meta.write_only_fields
    else:
        excluded = Meta.exclude | Meta.read_only_fields

    if hasattr(Meta, "model") and Meta.model is not None:
        for f, t in Meta.model.__fields__.items():
            if f not in excluded:
                f_def = t.default
                if t.required:
                    f_def = Required

                if t.shape == SHAPE_LIST:
                    _type = List[t.type_]
                elif t.shape == SHAPE_SET:
                    _type = Set[t.type_]
                elif t.shape == SHAPE_MAPPING:
                    _type = Mapping[t.key_field.type_, t.type_]
                elif t.shape == SHAPE_TUPLE:
                    _type = t.type_
                elif t.shape == SHAPE_TUPLE_ELLIPSIS:
                    _type = Tuple[t.type_, ...]
                elif t.shape == SHAPE_SEQUENCE:
                    _type = Sequence[t.type_]
                else:
                    _type = t.type_
                _fields.update({f: (_type, f_def)})

    for f, t in cls.__fields__.items():
        if f not in excluded:
            f_def = t.default
            if t.required:
                f_def = Required
            _fields.update({f: (t.type_, f_def)})

    if mode == FieldGenerationMode.REQUEST:
        response_model = gen_model(cls, mode=FieldGenerationMode.RESPONSE)

        Serializer.Config = Config
        model = create_model(cls.__name__, __base__=Serializer, **_fields)
        setattr(model, "response_model", response_model)
        setattr(model, "Meta", Meta)
        setattr(model, "Config", Config)

        reserved_attrs = ["Meta", "response_model", "Config"]
        for attr, value in cls.__dict__.items():
            if not attr.startswith("_") and attr not in reserved_attrs:
                setattr(model, attr, value)

        return model

    return create_model(f"{cls.__name__}Response", __config__=Config, **_fields)



import abc
from datetime import datetime
from enum import Enum
from typing import Any, List, Optional, Tuple

from pydantic import Field, root_validator, validator

from fa_common import CamelModel, DatabaseError, get_now
from fa_common import logger as LOG

from .utils import get_db_client

# from pymongo.results import UpdateResult, DeleteResult


class SortOrder(str, Enum):
    ASCENDING = "ASCENDING"
    DESCENDING = "DESCENDING"


class DBIndex(CamelModel):
    field_name: str


class FireOffset(CamelModel):
    # Start at or Start After
    start_at: bool = False
    start_id: Optional[str] = None
    start_fields: Optional[dict] = None

    @root_validator
    def check_passwords_match(cls, values):
        pw1, pw2 = values.get("start_id"), values.get("start_field")
        if pw1 is None and pw2 is None:
            raise ValueError("Either start_id or start_fields needs a value")
        return values


class Operator(str, Enum):
    EQUALS = "=="
    LT = "<"
    GT = ">"
    IN = "in"
    ARRAY_CONTAINS = "array_contains"
    ARRAY_CONTAINS_ANY = "array_contains_any"


class WhereCondition(CamelModel):
    field: str
    operator: Operator = Operator.EQUALS
    value: Any


class WriteResult(CamelModel):
    success: bool = True
    transform_results: Optional[List[Any]]


class DeleteResult(CamelModel):
    success: bool = True
    delete_time: Optional[datetime]


class DocumentDBModel(CamelModel, abc.ABC):
    """
    Base Model to use for any information saving in MongoDB.
    Provides `id` field as a base, populated by id-generator.
    Use it as follows:

    .. code-block:: python

        class MyModel(DocumentDBModel):
            additional_field1: str
            optional_field2: int = 42

            class Meta:
                collection = "mymodel_collection"

        mymodel = MyModel(additional_field1="value")
        mymodel.save()
        assert mymodel.additional_field1 == "value"
        assert mymodel.optional_field2 == 42
        assert isinstance(mymodel.id, int)
    """

    id: Optional[str] = Field(None, alias="_id")

    def set_id(self) -> str:
        """
        If id is supplied (ex. from DB) then use it, otherwise generate new.
        """
        if not self.id:
            db = get_db_client()
            self.id = db.generate_id(self.get_db_collection())
        if self.id is None:
            raise DatabaseError("Failed to generate a new id for {}", __name__)

        return self.id

    @classmethod
    @abc.abstractmethod
    def get_db_collection(cls) -> str:
        pass

    @classmethod
    async def get(cls, _id: str) -> Optional["DocumentDBModel"]:
        db = get_db_client()
        result = await db.get(cls, _id)
        return result

    @classmethod
    async def find_one(cls, where: List[WhereCondition]) -> Optional["DocumentDBModel"]:
        db = get_db_client()
        result = await db.find_one(cls, where)
        return result

    @classmethod
    async def delete(cls, _id: str) -> DeleteResult:
        db = get_db_client()
        result = await db.delete(cls, _id)
        return result

    @classmethod
    async def count(cls) -> int:
        db = get_db_client()
        result = await db.count(cls)
        return result

    @classmethod
    async def list(
        cls,
        where: List[WhereCondition] = [],
        _limit: int = 0,
        _sort: List[Tuple[str, SortOrder]] = None,
        mongo_offset: int = 0,
        fire_offset: FireOffset = None,
    ) -> List["DocumentDBModel"]:
        db = get_db_client()
        return await db.list(
            cls,
            where,
            _limit=_limit,
            _sort=_sort,
            mongo_offset=mongo_offset,
            fire_offset=fire_offset,
        )

    async def save(
        self, include: set = None, exclude: set = None, rewrite_fields: dict = None,
    ) -> str:
        db = get_db_client()

        _id = self.set_id()

        if not rewrite_fields:
            rewrite_fields = {}

        for field, value in rewrite_fields.items():
            setattr(self, field, value)

        insert_result = await db.insert(self, include=include, exclude=exclude)
        LOG.debug(insert_result.transform_results)
        return _id

    @classmethod
    async def update_one(cls, _id: str, data: dict) -> WriteResult:
        db = get_db_client()
        result = await db.update_one(cls, _id, data)
        return result

    @classmethod
    async def update_many(cls, data: List[Tuple[str, dict]]) -> WriteResult:
        db = get_db_client()
        result = await db.update_many(cls, data)
        return result

    @classmethod
    async def create_indexes(cls) -> Optional[List[str]]:
        if hasattr(cls, "Meta") and hasattr(cls.Meta, "indexes"):  # type:ignore
            db = get_db_client()
            return await db.create_indexes(cls, cls.Meta.indexes)  # type:ignore
        return None

    class Config:
        anystr_strip_whitespace = True


class DocumentDBTimeStampedModel(DocumentDBModel):
    """
    TimeStampedModel to use when you need to have `created` field,
    populated at your model creation time.

    Use it as follows:

    .. code-block:: python

        class MyTimeStampedModel(MongoDBTimeStampedModel):

            class Meta:
                collection = "timestamped_collection"


        mymodel = MyTimeStampedModel()
        mymodel.save()

        assert isinstance(mymodel.id, int)
        assert isinstance(mymodel.created, datetime)
    """

    created: Optional[datetime] = None

    @validator("created", pre=True, always=True)
    def set_created_now(cls, v: datetime) -> datetime:
        """
        If created is supplied (ex. from DB) -> use it, otherwise generate new.
        """
        if v:
            return v
        return get_now()



from typing import Optional

from fa_common import CamelModel, sizeof_fmt


# Shared properties
class File(CamelModel):
    size: Optional[str]  # e.g. '3 KB'
    size_bytes: Optional[int]
    url: Optional[str] = None  # download url
    gs_uri: Optional[str] = None  # GSC Uri
    id: Optional[str]  # id can be path or database id
    dir: bool = False
    path: Optional[str]  # path to current item (e.g. /folder1/someFile.txt)
    # optional (but we are using id as name if name is not present) (e.g. someFile.txt)
    name: str
    content_type: Optional[str]

    def set_size(self, bytes: int):
        self.size = sizeof_fmt(bytes)
        self.size_bytes = bytes



# import click
# import sys
# from loguru import logger
# from fa_common

# @click.group()
# @logger.catch
# def cli():
#     logger.remove()

#     logger.add(
#         sys.stdout,
#         enqueue=True,
#         level="DEBUG",
#         format="<green>{time:HH:mm:ss}</green> | <cyan>{process}</cyan> | <level>{message}</level>",
#     )


# @cli.command()
# @click.argument("job_id", type=str)
# @click.argument("filter_file_path", type=click.Path(exists=True))
# @click.option(
#     "--from", "-f", "from_", default="From", help="Filter Range From Column", type=str
# )
# @click.option("--to", "-t", default="To", help="Filter Range To Column", type=str)
# @click.option("--column", "-c", default="Depth", help="Data Filter Column", type=str)
# @click.option(
#     "--output", "-o", default="output.xlsx", help="File to output to", type=click.Path()
# )
# def job(data_file_path, filter_file_path, from_, to, column, output):
#     logger.debug("Filtering data from {} by range", data_file_path)
#     logger.debug("Filter File {}", filter_file_path)
#     logger.debug("Filter Range Columns From: {} To: {}", from_, to)
#     logger.debug("Data File Filter Column: {}", column)

#     filter_data_by_ranges_from_file(
#         data_file_path, filter_file_path, from_, to, column, output
#     )
#     logger.debug("Outputted to: {}", output)


# if __name__ == "__main__":
#     cli()



from fastapi import FastAPI

from fa_common import get_current_app, get_settings, logger
import oyaml as yaml


def setup_gitlab(app: FastAPI) -> None:
    """
    Helper function to setup MongoDB connection & `motor` client during setup.
    Use during app startup as follows:

    .. code-block:: python

        app = FastAPI()

        @app.on_event('startup')
        async def startup():
            setup_mongodb(app)

    :param app: app object, instance of FastAPI
    :return: None
    """
    settings = get_settings()
    if (
        settings.GITLAB_PRIVATE_TOKEN is not None
        and settings.GITLAB_URL is not None
        and settings.GITLAB_GROUP_ID is not None
    ):
        import gitlab

        gl = gitlab.Gitlab(
            settings.GITLAB_URL, private_token=settings.GITLAB_PRIVATE_TOKEN
        )
        app.gitlab = gl  # type:ignore
        logger.info("Gitlab client has been setup")

    else:
        raise ValueError(
            "Insufficient configuration to create gitlab client need (GITLAB_URL, GITLAB_PRIVATE_TOKEN and GITLAB_GROUP_ID)."
        )


def get_workflow_client():
    """
    Gets instance of GitlabClient for you to make gitlab calls.
    :return: GitlabClient
    """
    try:
        app = get_current_app()
        if app.gitlab is not None:
            from .client import GitlabClient

            client = GitlabClient()
            return client
    except Exception:
        raise ValueError("Problem returning gitlab client, may not be initialised.")
    raise ValueError("Gitlab client has not been initialised.")


class GitlabCIYAMLDumper(yaml.Dumper):
    """Correctly dumps yaml for gitlab ci formatting

    Arguments:
        yaml {[type]} -- [description]
    """

    def increase_indent(self, flow=False, indentless=False):
        return super(GitlabCIYAMLDumper, self).increase_indent(flow, False)



from fastapi import FastAPI
from fastapi.exceptions import RequestValidationError
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.requests import Request

from .responses import UJSONResponse
from .utils import logger as LOG


async def http_exception_handler(
    request: Request, exc: StarletteHTTPException
) -> UJSONResponse:
    """
    Handles StarletteHTTPException, translating it into flat dict error data:
        * code - unique code of the error in the system
        * detail - general description of the error
        * fields - list of dicts with description of the error in each field

    :param request: Starlette Request instance
    :param exc: StarletteHTTPException instance
    :return: UJSONResponse with newly formatted error data
    """
    fields = getattr(exc, "fields", [])
    data = {
        "code": getattr(exc, "error_code", exc.status_code),
        "detail": getattr(exc, "message", exc.detail),
        "fields": fields,
    }
    return UJSONResponse(data, status_code=exc.status_code)


async def validation_exception_handler(
    request: Request, exc: RequestValidationError
) -> UJSONResponse:
    """
    Handles ValidationError, translating it into flat dict error data:
        * code - unique code of the error in the system
        * detail - general description of the error
        * fields - list of dicts with description of the error in each field

    :param request: Starlette Request instance
    :param exc: StarletteHTTPException instance
    :return: UJSONResponse with newly formatted error data
    """
    LOG.warning(f"RequestValidationError Caught: {str(exc)}")
    status_code = getattr(exc, "status_code", 422)
    data = {
        "code": getattr(exc, "error_code", status_code),
        "detail": str(exc),
    }
    return UJSONResponse(data, status_code=status_code)


async def assert_exception(request: Request, exc: AssertionError) -> UJSONResponse:
    LOG.warning(f"Assert exception caught: {str(exc)}")
    status_code = getattr(exc, "status_code", 400)
    data = {
        "code": status_code,
        "detail": getattr(exc, "message", str(exc)),
    }
    return UJSONResponse(data, status_code=status_code)


async def default_exception(request: Request, exc: Exception) -> UJSONResponse:
    LOG.error(f"Internal exception caught: {str(exc)}")
    status_code = getattr(exc, "status_code", 500)
    data = {
        "code": status_code,
        "detail": getattr(exc, "message", "Internal server error"),
    }
    return UJSONResponse(data, status_code=status_code)


def setup_exception_handlers(app: FastAPI) -> None:
    """
    Helper function to setup exception handlers for app.
    Use during app startup as follows:

    .. code-block:: python

        app = FastAPI()

        @app.on_event('startup')
        async def startup():
            setup_exception_handlers(app)

    :param app: app object, instance of FastAPI
    :return: None
    """
    app.add_exception_handler(StarletteHTTPException, http_exception_handler)
    app.add_exception_handler(RequestValidationError, validation_exception_handler)
    app.add_exception_handler(AssertionError, assert_exception)
    app.add_exception_handler(Exception, default_exception)



import asyncio
import datetime
from io import BytesIO
from typing import List, Optional, Union

from fastapi import UploadFile
from google.cloud import storage
from google.cloud.exceptions import Conflict, GoogleCloudError, NotFound
from google.cloud.storage.bucket import Blob, Bucket

from fa_common import (
    StorageError,
    FileDownloadRef,
    force_async,
    get_current_app,
    get_settings,
    logger as LOG,
    sizeof_fmt,
)

from .base_client import BaseClient
from .model import File


class GoogleStorageClient(BaseClient):
    """
    Singleton client for interacting with GoogleStorage. Note we are wrapping all the call in threads to
    enable async support to a sync library.
    Please don't use it directly, use `core.storage.utils.get_storage_client`.
    """

    __instance = None
    gcp_storage: storage.Client = None

    def __new__(cls) -> "GoogleStorageClient":
        """
        Get called before the constructor __init__ and allows us to return a singleton instance.

        Returns:
            [GoogleStorageClient] -- [Singleton Instance of client]
        """
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
            app = get_current_app()
            cls.__instance.gcp_storage = app.gcp_storage  # type: ignore
        return cls.__instance

    async def make_bucket(self, name: str) -> None:
        try:
            await force_async(self.gcp_storage.create_bucket)(name)
        except Conflict:
            LOG.warning(f"Bucket {name} already exists")

    async def bucket_exists(self, name: str) -> bool:
        bucket = await force_async(self.gcp_storage.lookup_bucket)(name)
        if bucket is None:
            return False
        return True

    async def delete_bucket(self, name: str):
        try:
            bucket = await self._get_bucket(name)
            return await force_async(bucket.delete)(force=True)
        except ValueError:
            LOG.warning(
                f"Too many object in the bucket, unable to delete Bucket: {name}"
            )
            raise StorageError(
                f"Too many object in the bucket, unable to delete Bucket: {name}"
            )

    async def _get_bucket(self, name: str) -> Bucket:
        """
        Internal method to GCP bucket
        """
        try:
            return await force_async(self.gcp_storage.get_bucket)(name)
        except NotFound:
            LOG.error("Trying to get bucket {} that doesn't exist", name)
            raise StorageError(f"Trying to get bucket {name} that doesn't exist")

    async def _get_blob(self, bucket_name, path):
        """
        Internal method to get GCP blob reference
        """
        # bucket = self.gcp_storage.bucket(bucket_name)
        bucket = await self._get_bucket(bucket_name)
        return bucket.blob(self.convert_path_in(path, bucket_name))

    async def _list_blobs(self, bucket_name, prefix) -> List[Blob]:
        bucket = await self._get_bucket(bucket_name)
        return await force_async(bucket.list_blobs)(
            prefix=self.convert_path_in(prefix, bucket_name)
        )

    @classmethod
    def convert_path_in(cls, path: str, bucket_name) -> str:
        return path

    @classmethod
    def convert_path_out(cls, path: str, bucket_name) -> str:
        return path

    @classmethod
    def blob_to_file(cls, blob: Blob, bucket_name: str) -> Optional[File]:
        if blob.size is None and blob.exists():
            blob.reload()

        is_dir = blob.name.endswith("/")
        path = cls.convert_path_out(blob.name, bucket_name)
        path_segments = path.split("/")

        gs_uri = None
        if is_dir:
            if len(path_segments) == 1:
                return None
            path_segments = path_segments[0:-1]
        else:
            gs_uri = f"gs://{blob.bucket.name}/{blob.name}"

        name = path_segments[-1]
        path = "/".join(path_segments[0:-1])

        return File(
            id=blob.id,
            url=blob.public_url,
            gs_uri=gs_uri,
            size=sizeof_fmt(blob.size),
            size_bytes=blob.size,
            dir=is_dir,
            path=path,
            name=name,
            content_type=blob.content_type,
        )

    async def list_files(self, bucket_name: str, parent_path: str = "") -> List[File]:
        blobs = await self._list_blobs(bucket_name, parent_path)
        files: List[File] = []
        for blob in blobs:
            file = self.blob_to_file(blob, bucket_name)
            if file is not None:
                files.append(file)

        return files

    async def upload_string(
        self,
        string: Union[str, bytes],
        bucket_name: str,
        file_path: str,
        content_type="text/plain",
    ) -> File:

        blob = await self._get_blob(bucket_name, file_path)
        try:
            await force_async(blob.upload_from_string)(
                string, content_type=content_type
            )
        except GoogleCloudError as err:
            LOG.error(str(err))
            raise StorageError("Something went wrong uploading file {}", file_path)
        scidra_file = self.blob_to_file(blob, bucket_name)
        if scidra_file is None:
            raise StorageError("A file could not be created from the GCP blob")
        return scidra_file

    async def upload_file(
        self, file: UploadFile, bucket_name: str, parent_path: str = "",
    ) -> File:
        bucket = await self._get_bucket(bucket_name)
        path = self.convert_path_in(parent_path, bucket_name)
        if path != "":
            path += "/"

        blob = bucket.blob(path + file.filename)
        try:
            file.file.seek(0)
            await force_async(blob.upload_from_file)(
                file.file, content_type=file.content_type
            )
        except GoogleCloudError as err:
            LOG.error(str(err))
            raise StorageError(
                "Something went wrong uploadinf file {}", path + file.filename
            )
        scidra_file = self.blob_to_file(blob, bucket_name)
        if scidra_file is None:
            raise StorageError("A file could not be created from the GCP blob")
        return scidra_file

    async def get_file_ref(self, bucket_name: str, file_path: str) -> Optional[File]:
        blob = await self._get_blob(bucket_name, file_path)
        return self.blob_to_file(blob, bucket_name)

    async def get_file(self, bucket_name: str, file_path: str) -> Optional[BytesIO]:
        blob = await self._get_blob(bucket_name, file_path)
        if not blob or not blob.exists():
            return None

        byte_stream = BytesIO()
        await force_async(blob.download_to_file)(byte_stream)
        byte_stream.seek(0)
        return byte_stream

    async def file_exists(self, bucket_name: str, file_path: str) -> bool:
        blob = await self._get_blob(bucket_name, file_path)

        return blob.exists()

    async def folder_exists(self, bucket_name: str, path: str) -> bool:
        blobs = await self._list_blobs(bucket_name, path)
        for blob in blobs:
            return True

        return False

    async def _delete_blob(self, blob: Blob):
        await force_async(blob.delete)()

    async def delete_file(
        self, bucket_name: str, file_path: str, recursive: bool = False
    ) -> None:
        try:
            if recursive:
                blobs = await self._list_blobs(bucket_name, file_path)
                futures = [self._delete_blob(blob) for blob in blobs]
                await asyncio.wait(futures)
            else:
                blob = await self._get_blob(bucket_name, file_path)
                await self._delete_blob(blob)
        except NotFound:
            raise StorageError(
                f"Trying to delete a file {bucket_name}/{file_path} that doesn't exist."
            )

    async def rename_file(
        self, bucket_name: str, file_path: str, new_file_path: str
    ) -> None:
        bucket = await self._get_bucket(bucket_name)
        blob = bucket.blob(self.convert_path_in(file_path, bucket_name))
        if blob.exists():
            await force_async(bucket.rename_blob)(
                blob, self.convert_path_in(new_file_path, bucket_name)
            )
        else:
            raise StorageError(
                f"Trying to rename a file {bucket_name}/{file_path} that doesn't exist."
            )

        LOG.debug(f"{file_path} renamed to {new_file_path}")

    async def copy_file(
        self, from_bucket: str, from_path: str, to_bucket: str, to_path: str
    ) -> None:
        bucket = await self._get_bucket(from_bucket)
        blob = bucket.blob(self.convert_path_in(from_path, from_bucket))
        if blob.exists():
            await force_async(bucket.copy_blob)(
                blob,
                await self._get_bucket(to_bucket),
                self.convert_path_in(to_path, to_bucket),
            )
        else:
            raise StorageError(
                f"Trying to copy a file {from_bucket}/{from_path} that doesn't exist."
            )
        LOG.debug(f"{from_path} copied to {to_path}")

    async def create_temp_file_url(
        self, bucket: str, path: str, expire_time_hours: int = 3
    ) -> File:
        blob = await self._get_blob(bucket, path)
        expire_time = datetime.timedelta(hours=expire_time_hours)

        url = blob.generate_signed_url(expiration=expire_time, version="v4")
        LOG.info(f"Created url: {url} for file download")
        file_ref = self.blob_to_file(blob, bucket)
        if file_ref is None:
            raise StorageError(
                f"Cannot generate a temporary url {bucket}/{path} for a folder."
            )
        file_ref.url = url
        return file_ref


class FirebaseStorageClient(GoogleStorageClient):
    """
    Singleton client for interacting with FirebaseStorage. Note this is the same as the GoogleStorageClient
    except it uses a single bucket with folders
    Please don't use it directly, use `core.storage.utils.get_storage_client`.
    """

    __instance = None
    # gcp_storage: storage.Client = None
    bucket: storage.Bucket = None

    def __new__(cls) -> "FirebaseStorageClient":
        """
        Get called before the constructor __init__ and allows us to return a singleton instance.

        Returns:
            [FirebaseStorageClient] -- [Singleton Instance of client]
        """
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
            app = get_current_app()
            cls.__instance.gcp_storage = app.gcp_storage  # type: ignore
            cls.__instance.bucket = app.gcp_storage.get_bucket(get_settings().BUCKET_NAME)  # type: ignore
        return cls.__instance

    @classmethod
    def convert_path_in(cls, path: str, bucket_name) -> str:
        if path.startswith("/"):
            path = path[1:]
        return get_settings().BUCKET_USER_FOLDER + bucket_name + "/" + path

    @classmethod
    def convert_path_out(cls, path: str, bucket_name) -> str:
        return path.replace(get_settings().BUCKET_USER_FOLDER + bucket_name + "/", "")

    # Override to return single bucket for all users
    async def _get_bucket(self, name: str) -> Bucket:
        return self.bucket

    async def make_bucket(self, name: str) -> None:
        blob = self.bucket.blob(get_settings().BUCKET_USER_FOLDER + name + "/")
        await force_async(blob.upload_from_string)(
            "", content_type="application/x-www-form-urlencoded;charset=UTF-8"
        )

    async def bucket_exists(self, name: str) -> bool:
        blob = self.bucket.blob(get_settings().BUCKET_USER_FOLDER + name + "/")
        return await force_async(blob.exists)()

    @staticmethod
    def log_error_not_found(blob):
        LOG.warning("Trying to delete file {} that doesn't exist", blob.name)

    async def delete_bucket(self, name: str):
        blobs = await force_async(self.bucket.list_blobs)(
            prefix=get_settings().BUCKET_USER_FOLDER + name + "/"
        )
        await force_async(self.bucket.delete_blobs)(
            blobs, on_error=self.log_error_not_found
        )



from io import BytesIO
from typing import List, Optional, Union

from fastapi import UploadFile
from minio import Minio, Object
from minio.error import (
    BucketAlreadyExists,
    BucketAlreadyOwnedByYou,
    NoSuchKey,
    ResponseError,
)

from fa_common import StorageError, force_async, get_current_app, FileDownloadRef
from fa_common import logger as LOG
from fa_common import sizeof_fmt

from .base_client import BaseClient
from .model import File


class MinioClient(BaseClient):
    """
    Singleton client for interacting with Minio. Note we are wrapping all the call in threads to
    enable async support to a sync library.
    Please don't use it directly, use `core.storage.utils.get_storage_client`.
    """

    __instance = None
    minio: Minio = None

    def __new__(cls) -> "MinioClient":
        """
        Get called before the constructor __init__ and allows us to return a singleton instance.

        Returns:
            [MinioClient] -- [Singleton Instance of client]
        """
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
            app = get_current_app()
            cls.__instance.minio = app.minio  # type: ignore
        return cls.__instance

    async def make_bucket(self, name: str) -> None:
        try:
            await force_async(self.minio.make_bucket)(name)
        except BucketAlreadyOwnedByYou:
            LOG.warning(f"Bucket {name} already owned by app")
        except BucketAlreadyExists:
            LOG.warning(f"Bucket {name} already exists")
        except ResponseError as err:
            LOG.error(f"Unable to create bucket {name}")
            LOG.error(err)
            raise err

    async def bucket_exists(self, name: str) -> bool:
        return await force_async(self.minio.bucket_exists)(name)

    @classmethod
    def object_to_file(
        cls, obj: Object, bucket_name: str, file_name: str = None
    ) -> Optional[File]:
        is_dir = obj.is_dir
        path = obj.object_name
        path_segments = path.split("/")

        if is_dir:
            if len(path_segments) == 1:
                return None
            path_segments = path_segments[0:-1]

        name = path_segments[-1]
        path = "/".join(path_segments[0:-1])

        LOG.debug("Converting Minio Object: {}", obj)

        return File(
            id=obj.object_name,
            url=obj.bucket_name + obj.object_name,
            size=sizeof_fmt(obj.size),
            size_bytes=obj.size,
            dir=is_dir,
            path=path,
            name=name,
            content_type=obj.content_type,
        )

    async def list_files(self, bucket_name: str, parent_path: str = "") -> List[File]:
        objects = await force_async(self.minio.list_objects_v2)(
            bucket_name, prefix=parent_path
        )
        files: List[File] = []
        for obj in objects:
            file = self.object_to_file(obj, bucket_name)
            if file is not None:
                files.append(file)
        return files

    async def upload_string(
        self,
        string: Union[str, bytes],
        bucket_name: str,
        file_path: str,
        content_type="text/plain",
    ) -> File:
        try:
            string_io: BytesIO = (
                BytesIO(string.encode("utf-8"))
                if isinstance(string, str)
                else BytesIO(string)
            )
            await force_async(self.minio.put_object)(
                bucket_name,
                file_path,
                string_io,
                len(string),
                content_type=content_type,
            )

        except ResponseError as err:
            LOG.error(str(err))
            raise StorageError("Something went wrong uploading file {}", file_path)
        obj = await force_async(self.minio.stat_object)(bucket_name, file_path)
        scidra_file = self.object_to_file(obj, bucket_name, file_path)
        if scidra_file is None:
            raise StorageError("A file could not be created from the Minio obj")
        return scidra_file

    async def upload_file(
        self, file: UploadFile, bucket_name: str, parent_path: str = "",
    ) -> File:

        if parent_path != "":
            parent_path += "/"
        try:
            await force_async(self.minio.fput_object)(
                bucket_name, parent_path + file.filename, file.file.fileno()
            )

        except ResponseError as err:
            LOG.error(str(err))
            raise StorageError(
                "Something went wrong uploading file {}", parent_path + file.filename
            )
        obj = await force_async(self.minio.stat_object)(
            bucket_name, parent_path + file.filename
        )
        scidra_file = self.object_to_file(obj, bucket_name, parent_path + file.filename)
        if scidra_file is None:
            raise StorageError("A file could not be created from the Minio obj")
        return scidra_file

    async def get_file_ref(self, bucket_name: str, file_path: str) -> Optional[File]:
        obj_ref = await force_async(self.minio.stat_object)(bucket_name, file_path)
        return self.object_to_file(obj_ref, bucket_name)

    async def get_file(self, bucket_name: str, file_path: str) -> Optional[BytesIO]:
        try:
            obj = await force_async(self.minio.get_object)(bucket_name, file_path)
            bytes_stream = BytesIO()
            for d in obj.stream(32 * 1024):
                bytes_stream.write(d)

            return bytes_stream
        except NoSuchKey:
            return None
        except ResponseError as err:
            # Likely the file doesn't exist
            LOG.error(str(err))
            raise StorageError("Error getting file {}", file_path)

    async def file_exists(self, bucket_name: str, file_path: str) -> bool:
        try:
            stats = await force_async(self.minio.stat_object)(bucket_name, file_path)
            LOG.info(stats)
            if stats is None:
                return False
            return True
        except NoSuchKey:
            return False
        except ResponseError as err:
            LOG.error(str(err))
            raise StorageError("Error checking if {} exists", file_path)

    async def folder_exists(self, bucket_name: str, path: str) -> bool:
        objects = await force_async(self.minio.list_objects_v2)(
            bucket_name, prefix=path
        )
        for obj in objects:
            if obj is not None:
                return True
        else:
            return False

    async def delete_file(
        self, bucket_name: str, file_path: str, recursive: bool = False
    ) -> None:
        try:
            if recursive:
                objects = await force_async(self.minio.list_objects_v2)(
                    bucket_name, prefix=file_path, recursive=True
                )
                obj_list = []
                for obj in objects:
                    obj_list.append(obj.object_name)
                if len(obj_list) > 0:
                    for del_err in self.minio.remove_objects(bucket_name, obj_list):
                        LOG.error(
                            f"Error while deleting objects in bucket {bucket_name}: {del_err}"
                        )
            else:
                await force_async(self.minio.remove_object)(bucket_name, file_path)
        except ResponseError as err:
            LOG.error(err)
            raise StorageError(
                f"Somthing went wrong deleting file(s) {bucket_name}/{file_path}"
            )

    async def delete_bucket(self, name: str):
        try:
            await self.delete_file(name, "", True)
            return await force_async(self.minio.remove_bucket)(name)
        except ResponseError as err:
            LOG.error(f"Unable to delete bucket {name}")
            LOG.error(err)
            raise err

    async def rename_file(
        self, bucket_name: str, file_path: str, new_file_path: str
    ) -> None:
        try:
            await force_async(self.minio.copy_object)(
                bucket_name, new_file_path, f"/{bucket_name}/{file_path}"
            )
        except ResponseError as err:
            LOG.error(
                f"Error renaming {file_path} to {new_file_path} in bucket {bucket_name}: {str(err)}"
            )
            raise StorageError(
                f"Somthing went wrong renaming {file_path} to {new_file_path}"
            )
        await self.delete_file(bucket_name, file_path)

    async def copy_file(
        self, from_bucket: str, from_path: str, to_bucket: str, to_path: str
    ) -> None:
        try:
            await force_async(self.minio.copy_object)(
                to_bucket, to_path, f"/{from_bucket}/{from_path}"
            )
        except ResponseError as err:
            LOG.error(
                f"Error copying {from_bucket} {from_path} to {to_bucket} {to_path} : {str(err)}"
            )
            raise StorageError(f"Somthing went wrong copying {from_path} to {to_path}")

    async def create_temp_file_url(
        self, bucket: str, path: str, expire_time_hours: int = 3
    ) -> File:
        # FIXME make this work
        raise NotImplementedError(
            "Minio support for temp downloads currently not working."
        )



from typing import List, Optional

from fastapi import Depends
from fastapi.openapi.models import OAuthFlows
from fastapi.security import OAuth2, SecurityScopes
from jose import JWTError, jwt
from pydantic import ValidationError

from fa_common import (
    ForbiddenError,
    UnauthorizedError,
    async_get,
    force_async,
    get_settings,
)
from fa_common import logger as LOG

from .models import AuthUser

oauth2_scheme = OAuth2(
    flows=OAuthFlows(
        implicit={
            "authorizationUrl": get_settings().OAUTH2_AUTH_URL,
            "scopes": {"openid": "", "profile": "", "email": ""},
        }
    )
)


async def get_firebase_token(user: AuthUser) -> str:
    from firebase_admin import auth as fire_auth

    uid = user.sub

    # Potentially add scopes here
    # additional_claims = {
    #     "premiumAccount": True
    # }
    additional_claims: dict = {}
    return await force_async(fire_auth.create_custom_token)(uid, additional_claims)


async def get_user_profile(url: str, auth: str) -> AuthUser:
    data = await async_get(url, auth)
    return AuthUser(**data)


def get_token_auth_header(auth: str) -> str:
    """Obtains the Access Token from the Authorization Header
    """
    if not auth:
        raise UnauthorizedError(detail="Authorization header is expected",)

    parts = auth.split()

    if parts[0].lower() != "bearer":
        raise UnauthorizedError(detail="Authorization header must start with Bearer",)
    elif len(parts) == 1:
        raise UnauthorizedError(detail="Token not found",)
    elif len(parts) > 2:
        raise UnauthorizedError(detail="Authorization header must be Bearer token",)

    token = parts[1]
    return token


async def decode_token(auth: str):
    settings = get_settings()
    token = get_token_auth_header(auth)
    jwks = await async_get(f"https://{settings.AUTH0_DOMAIN}/.well-known/jwks.json")

    unverified_header = jwt.get_unverified_header(token)
    rsa_key: dict = {}
    for key in jwks["keys"]:
        if key["kid"] == unverified_header["kid"]:
            rsa_key = {
                "kty": key["kty"],
                "kid": key["kid"],
                "use": key["use"],
                "n": key["n"],
                "e": key["e"],
            }
    if rsa_key:
        try:
            payload = jwt.decode(
                token,
                rsa_key,
                algorithms=settings.JWT_ALGORITHMS,
                audience=settings.API_AUDIENCE,
                issuer="https://" + settings.AUTH0_DOMAIN + "/",
            )
        except jwt.ExpiredSignatureError:
            raise UnauthorizedError(detail="token is expired",)
        except jwt.JWTClaimsError:
            raise UnauthorizedError(
                detail="incorrect claims, please check the audience and issuer",
            )
        except Exception:
            raise UnauthorizedError(detail="Unable to parse authentication token.",)

        return payload

    raise UnauthorizedError(detail="Unable to find appropriate key",)


async def get_user(payload: dict, token: str) -> Optional[AuthUser]:
    user: Optional[AuthUser] = None
    token_scopes: List[str] = []

    if "permissions" in payload and payload.get("permissions") is not None:
        token_scopes = payload["permissions"]

    if "sub" in payload:
        if (
            "aud" in payload
            and len(payload["aud"]) > 1
            and "userinfo" in payload["aud"][1]
        ):
            user = await get_user_profile(payload["aud"][1], token)
        else:
            user = AuthUser(**payload)

        if len(token_scopes) > 0:
            user.scopes = token_scopes

    return user


async def get_current_user(
    security_scopes: SecurityScopes, token: str = Depends(oauth2_scheme)
) -> AuthUser:

    try:
        payload = await decode_token(token)
        user = await get_user(payload, token)

    except (JWTError, ValidationError) as e:
        LOG.error("Could not validate credentials. {}", str(e))
        raise UnauthorizedError(detail="Could not validate credentials.",)

    if not user:
        raise UnauthorizedError(detail="Invalid authentication credentials.",)

    for scope in security_scopes.scopes:
        if scope not in user.scopes:
            raise ForbiddenError(detail="Not enough permissions",)
    return user



from .utils import get_workflow_client, setup_gitlab
from .service import (
    run_job,
    get_job_run,
    get_job_log,
    create_workflow_project,
    delete_workflow_project,
    delete_workflow_user,
)
from .models import (
    ScidraModule,
    WorkflowProject,
    WorkflowRun,
    JobRun,
    FileFieldDescription,
    JobStatus,
    ModuleType,
)



from typing import Any, List, Optional, Tuple, Type

from google.cloud.exceptions import NotFound
from google.cloud.firestore import Client, CollectionReference, Query

from fa_common import DatabaseError, force_async, get_current_app
from fa_common import logger as LOG

from .base_client import BaseClient
from .models import (
    DBIndex,
    DeleteResult,
    DocumentDBModel,
    FireOffset,
    SortOrder,
    WhereCondition,
    WriteResult,
)


class FirestoreClient(BaseClient):
    """
    Singleton client for interacting with Firestore.
    Operates mostly using models, specified when making DB queries.

    Implements only part of internal `motor` methods, but can be populated more.

    Please don't use it directly, use `scidra.core.db.utils.get_db_client`.
    """

    __instance = None
    firestore: Client = None

    def __new__(cls) -> "FirestoreClient":
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
            app = get_current_app()
            cls.__instance.firestore = app.firestore  # type: ignore
        return cls.__instance

    def get_collection(self, collection_name: str) -> CollectionReference:
        return self.firestore.collection(collection_name)

    def generate_id(self, collection_name: str) -> str:
        collection = self.get_collection(collection_name)
        return collection.document().id

    async def insert(
        self, model: DocumentDBModel, include=None, exclude=None, session: Any = None,
    ) -> WriteResult:
        data = model.dict(include=include, exclude=exclude)
        _id: str = data["id"]
        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)

        try:
            result = await force_async(collection.add)(data, _id)
        except Exception as err:
            LOG.error(f"Error caught adding record to firestore: {str(err)}")
            LOG.error(f"Dict = {data}")
            raise DatabaseError(f"Document with id: {_id} already exists")

        LOG.info("Insert Result: {}", result[1].id)

        return WriteResult(transform_results=[result[1].id])

    async def count(self, model: Type[DocumentDBModel], session: Any = None) -> int:

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        docs = await force_async(collection.list_documents)()

        return len(docs)

    async def delete(
        self, model: Type[DocumentDBModel], _id: str, session: Any = None
    ) -> DeleteResult:

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        doc_ref = collection.document(_id)
        try:
            timestamp = await force_async(doc_ref.delete)()
            LOG.info(f"User deleted at {timestamp}")
            return DeleteResult()
        except NotFound as exc:
            LOG.warning(f"Error deleting user {_id} from firestore, user not found")
            raise exc

    async def update_one(
        self, model: Type[DocumentDBModel], _id: str, data: dict, session: Any = None
    ) -> WriteResult:

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        await force_async(collection.document(_id).update)(data)
        return WriteResult()

    # FIXME: Intended to update all docs that meet search criteria with a single set of values
    # async def update_many(
    #     self, model: Type[DocumentDBModel], data: List[Tuple[str, dict]]
    # ) -> WriteResult:

    #     collection_name = model.get_db_collection()
    #     collection = self.get_collection(collection_name)

    #     batch = self.firestore.batch()
    #     for _id, _data in data:
    #         doc_ref = collection.document(_id)

    #         batch.update(doc_ref, field_updates=_data)
    #     await force_async(batch.commit)()

    #     return WriteResult()

    async def get(
        self, model: Type[DocumentDBModel], _id: str, session: Any = None
    ) -> Optional[DocumentDBModel]:

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)

        result = await force_async(collection.document(_id).get)()

        if not result.exists:
            return None

        return model(**result.to_dict())

    async def find_one(
        self,
        model: Type[DocumentDBModel],
        where: List[WhereCondition] = [],
        session: Any = None,
    ) -> Optional[DocumentDBModel]:

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        query_ref = collection
        for wc in where:
            query_ref = query_ref.where(wc.field, wc.operator, wc.value)
        query_ref.limit(1)

        result = await force_async(query_ref.stream)()

        for res in result:
            LOG.debug("Find One Result: {}", res.to_dict())
            item = model(**res.to_dict())
            item.id = res.id
            return item

        return None

    async def list(
        self,
        model: Type[DocumentDBModel],
        where: List[WhereCondition] = [],
        _limit: int = 0,
        _sort: List[Tuple[str, SortOrder]] = None,
        mongo_offset: int = 0,
        fire_offset: FireOffset = None,
    ) -> List[DocumentDBModel]:

        collection_name = model.get_db_collection()
        collection = self.get_collection(collection_name)
        offset_value = None
        if fire_offset is not None:
            if fire_offset.start_id is not None:
                offset_value = await force_async(
                    collection.document(fire_offset.start_id).get
                )()
            else:
                offset_value = fire_offset.start_fields
        query_ref = collection
        for wc in where:
            query_ref = query_ref.where(wc.field, wc.operator, wc.value)
        if _sort is not None:
            for s in _sort:
                direction = (
                    Query.DESCENDING if s is SortOrder.DESCENDING else Query.ASCENDING
                )
                query_ref = query_ref.order_by(s[0], direction=direction)
        if offset_value is not None:
            if fire_offset.start_at:  # type: ignore
                query_ref = query_ref.start_at(offset_value)
            else:
                query_ref = query_ref.start_after(offset_value)
        if _limit > 0:
            query_ref = query_ref.limit(_limit)

        results = await force_async(query_ref.stream)()

        models: List[DocumentDBModel] = []
        for res in results:

            models.append(model(**res.to_dict()))
        return models

    async def create_indexes(self, index=List[DBIndex]) -> Optional[List[str]]:
        return None



import importlib
import random
from pathlib import Path
from typing import List

from fastapi import FastAPI

from fa_common import DatabaseType, get_settings, logger


def default_id_generator(bit_size: int = 32) -> int:
    """
    Generator of IDs for newly created MongoDB rows.

    :return: `bit_size` long int
    """
    return random.getrandbits(bit_size)


def get_next_id() -> int:
    """
    Retrieves ID generator function from the path, specified in project's conf.
    :return: newly generated ID
    """
    return default_id_generator()


def setup_db(app: FastAPI) -> None:
    """
    Helper function to setup MongoDB connection & `motor` client during setup.
    Use during app startup as follows:

    .. code-block:: python

        app = FastAPI()

        @app.on_event('startup')
        async def startup():
            setup_mongodb(app)

    :param app: app object, instance of FastAPI
    :return: None
    """
    settings = get_settings()
    if settings.DATABASE_TYPE == DatabaseType.MONGODB:
        # Only import mongo deps if we are using mongo
        import motor.motor_asyncio

        client = motor.motor_asyncio.AsyncIOMotorClient(
            settings.MONGODB_DSN,
            minPoolSize=settings.mongodb_min_pool_size,
            maxPoolSize=settings.mongodb_max_pool_size,
        )
        app.mongodb = client[settings.MONGODB_DBNAME]  # type: ignore
        logger.info("Mongo Database has been set")

    elif settings.DATABASE_TYPE == DatabaseType.GCP_FIRESTORE:
        from firebase_admin import firestore

        app.firestore = firestore.client()  # type: ignore
        logger.info("Firestore client has been set")
    elif settings.DATABASE_TYPE == DatabaseType.NONE:
        logger.info("Database is set to NONE and cannot be used")
        return
    else:
        raise ValueError("DATABASE_TYPE Setting is not a valid database option.")


def get_db_client():
    """
    Gets instance of BaseClient client for you to make DB queries.
    :return: BaseClient
    """
    settings = get_settings()
    if settings.DATABASE_TYPE == DatabaseType.MONGODB:
        from .mongo_client import MongoDBClient

        client = MongoDBClient()
        return client
    elif settings.DATABASE_TYPE == DatabaseType.GCP_FIRESTORE:
        from .firestore_client import FirestoreClient

        client = FirestoreClient()
        return client

    raise ValueError("DATABASE_TYPE Setting is not a valid database option.")


def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)]
    )


def get_models() -> list:
    """
    Scans `settings.APP_PATH`.
    Find `models` modules in each of them and get all attributes there.
    Last step is to filter attributes to return only those,
    subclassed from DocumentDBModel (or timestamped version).

    Used internally only by `create_indexes` function.

    :return: list of user-defined models (subclassed from DocumentDBModel) in apps
    """
    from .models import DocumentDBModel

    models = Path(get_settings().APP_PATH).glob("**/models.py")
    for m in models:
        mod_string = str(m).replace("/", ".").replace("\\", ".").replace(".py", "")
        importlib.import_module(mod_string)

    return list(all_subclasses(DocumentDBModel))


async def create_indexes() -> List[str]:
    """
    Gets all models in project and then creates indexes for each one of them.
    :return: list of indexes that has been invoked to create
             (could've been created earlier, it doesn't raise in this case)
    """
    models = get_models()
    indexes = []
    for model in models:
        indexes.append(await model.create_indexes())
    return list(filter(None, indexes))



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="fa-common",
    description="FastAPI Common Framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version="0.8.1",
    packages=find_packages(),
    maintainer="Samuel Bradley",
    maintainer_email="sam.bradley@csiro.au",
    python_requires=">=3.5, <4",
    # entry_points={"console_scripts": ["geo-data-utils=fa_common.__cli__:cli"]},
    install_requires=[
        "pyjwt",
        "python-jose",
        "python-dotenv",
        "loguru",
        "python-multipart",
        "email-validator",
        "six",
        "fastapi>=0.52.0",
        "aiohttp",
        "ujson",
        "humps",
        "click",
    ],
    extras_require={
        "sentry": ["sentry_sdk"],
        "mongo": ["motor", "pymongo"],
        "gcp": [
            "google-cloud-storage",
            "google-cloud-logging",
            "google-cloud-firestore",
            "firebase-admin",
            "requests",
        ],
        "minio": ["minio"],
        "windows": ["win32-setctime"],
        "gitlab": ["oyaml", "python-gitlab"],
    },
    package_data={"fa_common.workflow": ["*.yml"]},
    include_package_data=True,
    license="MIT",
    url="https://gitlab.com/csiro-geoanalytics/python-shared/fastapi-common-framework",
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Science/Research",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



import os
import json
from datetime import datetime
from typing import List
from .models import ScidraModule, JobRun, WorkflowProject
from .utils import get_workflow_client
from fa_common import logger as LOG
from fa_common.storage import File
import oyaml as yaml

dirname = os.path.dirname(__file__)


async def create_workflow_project(user_id: str, project_name: str) -> WorkflowProject:
    client = get_workflow_client()
    try:
        project = await client._get_project_by_name(user_id)
    except ValueError:
        LOG.info(f"Workflow User {user_id} does not exist, creating.")
        project = await client.create_project(user_id)

    branch = await client.create_branch(project.id, project_name)

    return WorkflowProject(
        name=branch.name,
        user_id=user_id,
        gitlab_project_id=project.id,
        created=str(datetime.now()),
    )


async def delete_workflow_project(user_id: str, project_name: str):
    client = get_workflow_client()

    try:
        await client.delete_branch(user_id, project_name)
    except ValueError:
        LOG.error(
            f"Trying to delete workflow project {project_name} does not exist for user {user_id}."
        )
        raise ValueError(f"Workflow Project {project_name} does not exist.")


async def delete_workflow_user(user_id: str, wait: bool = False):
    client = get_workflow_client()
    try:
        await client.delete_project_by_name(user_id, wait)
    except ValueError:
        LOG.error(f"Workflow User {user_id} does not exist.")
        raise ValueError(f"Workflow User {user_id} does not exist.")


async def run_job(
    project: WorkflowProject,
    description: str,
    module: ScidraModule,
    job_data: dict,
    runner: str = "csiro-swarm",
    files: List[File] = [],
    sync: bool = False,
) -> JobRun:

    file_refs = []
    for _file in files:
        file_refs.append(_file.dict())

    with open(os.path.join(dirname, "job.yml")) as yaml_file:
        job_yaml = yaml.safe_load(yaml_file)

    job_yaml["run-job"]["tags"] = [runner]
    job_yaml["run-job"]["image"] = module.docker_image
    job_yaml["run-job"]["variables"]["TZ"] = project.timezone
    job_yaml["run-job"]["variables"]["JOB_PARAMETERS"] = json.dumps(job_data)
    job_yaml["run-job"]["variables"]["FILE_REFS"] = json.dumps(file_refs)
    job_yaml["run-job"]["variables"]["MODULE_NAME"] = module.name
    job_yaml["run-job"]["variables"]["KUBERNETES_CPU_REQUEST"] = module.cpu_request
    job_yaml["run-job"]["variables"]["KUBERNETES_CPU_LIMIT"] = module.cpu_limit
    job_yaml["run-job"]["variables"][
        "KUBERNETES_MEMORY_REQUEST"
    ] = f"{module.memory_request_gb}Gi"
    job_yaml["run-job"]["variables"][
        "KUBERNETES_MEMORY_LIMIT"
    ] = f"{module.memory_limit_gb}Gi"

    client = get_workflow_client()

    await client.update_ci(
        project.gitlab_project_id, project.name, job_yaml, description
    )
    job_run = await client.run_pipeline(
        project.gitlab_project_id, project.name, wait=sync
    )
    return job_run.jobs[0]


async def get_job_run(user_id: str, job_id: int, include_log: bool = False) -> JobRun:
    client = get_workflow_client()
    return await client.get_job(user_id, job_id, include_log)


async def get_job_log(user_id: str, job_id: int) -> bytes:
    client = get_workflow_client()
    logs = await client.get_job_log(user_id, job_id)
    return logs



import logging
import os
from enum import Enum
from typing import List, Optional, Set

from dotenv import load_dotenv
from pydantic import AnyUrl, BaseSettings, SecretBytes, SecretStr

# from app.shared.utils import auth


class StorageType(str, Enum):
    MINIO = "MINIO"
    GCP_STORAGE = "GCP_STORAGE"
    FIREBASE_STORAGE = "FIREBASE_STORAGE"
    NONE = "NONE"


class DatabaseType(str, Enum):
    MONGODB = "MONGODB"
    GCP_FIRESTORE = "GCP_FIRESTORE"
    NONE = "NONE"


class Settings(BaseSettings):
    VERSION: str = "0.8.1"
    API_VERSION: int = 1
    API_PRE_PATH: str = f"/api/v{API_VERSION}"

    SENTRY_DSN: Optional[AnyUrl] = None
    DEBUG: bool = False
    UNIT_TESTING: bool = False

    SECRET_KEY: SecretBytes = os.urandom(32)  # type:ignore
    BUILD_DATE: Optional[str] = None
    PROJECT_NAME: str = "FastAPI Backend"
    BACKEND_CORS_ORIGINS: Set[str] = set()
    MINIO_SECRET_KEY: Optional[SecretStr] = None
    MINIO_ACCESS_KEY: Optional[str] = None
    MINIO_ENDPOINT: Optional[str] = None
    MINIO_SSL: bool = False

    GITLAB_PRIVATE_TOKEN: Optional[str] = None
    GITLAB_GROUP_ID: Optional[int] = None
    GITLAB_URL: Optional[str] = "https://gitlab.com/"

    BUCKET_NAME: str = ""
    BUCKET_PREFIX: str = ""
    BUCKET_USER_FOLDER: str = "user-storage/"

    AUTH0_DOMAIN: str = ""
    API_AUDIENCE: str = ""
    OAUTH2_AUTH_URL: str = f"https://{AUTH0_DOMAIN}/authorize?audience={API_AUDIENCE}"
    JWT_ALGORITHMS: List[str] = ["RS256"]

    # logging configuration
    LOGGING_LEVEL: int = logging.DEBUG if DEBUG else logging.INFO

    DATABASE_TYPE: DatabaseType = DatabaseType.GCP_FIRESTORE  # FIRESTORE or MONGODB
    STORAGE_TYPE: StorageType = StorageType.GCP_STORAGE
    USE_FIREBASE: bool = (
        STORAGE_TYPE == StorageType.FIREBASE_STORAGE
        or DATABASE_TYPE == DatabaseType.GCP_FIRESTORE
    )

    MONGODB_DSN: Optional[str] = None
    MONGODB_DBNAME: Optional[str] = None
    mongodb_min_pool_size: int = 0
    mongodb_max_pool_size: int = 100
    TZ: str = "Australia/Perth"
    APP_PATH: str = "fa_common"
    FASTAPI_APP: str = ""

    debug_timing: bool = False

    # class Config:
    #     env_prefix = 'FA_COMMON_'  # defaults to no prefix, i.e. ""
    #     fields = {
    #         'auth_key': {
    #             'env': 'my_auth_key',
    #         },
    #         'redis_dsn': {
    #             'env': ['service_redis_dsn', 'redis_url']
    #         }
    #     }


settings: Optional[Settings] = None


def get_settings(env_path=None) -> Settings:
    # Load env variables from .env file
    if env_path is not None:
        load_dotenv(dotenv_path=env_path)

    global settings
    if settings is None or env_path is not None:
        settings = Settings()

    return settings



from typing import Any, Dict, List

from starlette.exceptions import HTTPException as StarletteHTTPException


class DatabaseError(Exception):
    """
    Simple exception to notify something unexpected has happened with a
    database call
    """


class StorageError(Exception):
    """
    Simple exception to notify something unexpected has happened with a
    storage call
    """


class HTTPException(StarletteHTTPException):
    def __init__(
        self,
        status_code: int,
        error_code: int,
        detail: Any = None,
        fields: List[Dict] = None,
    ) -> None:
        """
        Generic HTTP Exception with support for custom status & error codes.

        :param status_code: HTTP status code of the response
        :param error_code: Custom error code, unique throughout the app
        :param detail: detailed message of the error
        :param fields: list of dicts with key as field and value as message
        """
        super().__init__(status_code=status_code, detail=detail)
        self.error_code = error_code
        self.fields = fields or []


class NotImplementedError(HTTPException):
    def __init__(self, detail: Any, fields: List[Dict] = None):
        """
        Generic Not implemented error

        :param detail: detailed message of the error
        """
        super().__init__(
            error_code=501, status_code=501, detail=detail, fields=fields,
        )


class UnknownError(HTTPException):
    def __init__(self, detail: Any, fields: List[Dict] = None):
        """
        Generic Unknown error that has been explicitly caught, uses "BadRequest" status code

        :param detail: detailed message of the error
        """
        super().__init__(
            error_code=4001, status_code=400, detail=detail, fields=fields,
        )


class BadRequestError(HTTPException):
    def __init__(self, error_code: int, detail: Any, fields: List[Dict] = None):
        """
        Generic Bad Request HTTP Exception with support for custom error code.

        :param error_code: Custom error code, unique throughout the app
        :param detail: detailed message of the error
        """
        super().__init__(
            error_code=error_code, status_code=400, detail=detail, fields=fields,
        )


class UnauthorizedError(HTTPException):
    def __init__(
        self,
        error_code: int = 401,
        detail: Any = "Unauthorized",
        fields: List[Dict] = None,
    ):
        """
        Generic Unauthorized HTTP Exception with support for custom error code.

        :param error_code: Custom error code, unique throughout the app
        :param detail: detailed message of the error
        """
        super().__init__(
            error_code=error_code, status_code=401, detail=detail, fields=fields,
        )


class ForbiddenError(HTTPException):
    def __init__(
        self,
        error_code: int = 403,
        detail: Any = "Forbidden.",
        fields: List[Dict] = None,
    ):
        """
        Generic Forbidden HTTP Exception with support for custom error code.

        :param error_code: Custom error code, unique throughout the app
        :param detail: detailed message of the error
        """
        super().__init__(
            error_code=error_code, status_code=403, detail=detail, fields=fields,
        )


class NotFoundError(HTTPException):
    def __init__(
        self,
        error_code: int = 404,
        detail: Any = "Not found.",
        fields: List[Dict] = None,
    ):
        """
        Generic 404 Not Found HTTP Exception with support for custom error code.

        :param error_code: Custom error code, unique throughout the app
        :param detail: detailed message of the error
        """
        super().__init__(
            error_code=error_code, status_code=404, detail=detail, fields=fields,
        )



from enum import Enum
from typing import Optional, List, Union
import pytz

from pydantic import validator
from fa_common import CamelModel

# from fa_common.db import DocumentDBModel


class ModuleType(str, Enum):
    SYNC = "sync"  # Is run via a service call
    ASYNC = "async"  # Is executed via gitlab ci


class JobStatus(str, Enum):
    NOT_SET = ""
    RECEIVED = "RECEIVED"
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    COMPLETE = "COMPLETE"
    FAILED = "FAILED"


class FileFieldDescription(CamelModel):
    name: str
    description: str
    valid_extensions: List[str]
    max_size: Optional[int]
    mandatory: bool = False


class ScidraModule(CamelModel):
    version: str = "1.0.0"
    name: str
    description: str = ""
    module_type: ModuleType = ModuleType.ASYNC
    docker_image: str
    input_schema: str = ""
    output_schema: str = ""
    input_files: List[FileFieldDescription] = []
    cpu_limit: int = 4
    cpu_request: int = 1
    memory_limit_gb: int = 8
    memory_request_gb: int = 2


class JobRun(CamelModel):
    id: int
    status: str = ""
    started_at: Optional[str]
    finished_at: Optional[str]
    duration: Optional[int]
    name: Optional[str]
    stage: Optional[str]
    output: Optional[Union[List, dict]]
    # output_files:
    log: Optional[bytes]


class WorkflowRun(CamelModel):
    """Equivilant to  gitlab pipeline"""

    id: int
    gitlab_project_id: int
    gitlab_project_branch: str
    commit_id: str
    status: str = ""
    jobs: List[JobRun] = []
    started_at: Optional[str]
    finished_at: Optional[str]
    duration: Optional[int]


# class ScidraJob(CamelModel):
#     id: int
#     module_id: str
#     status: Optional[JobStatus]
#     inputs: dict = {}
#     outputs: dict = {}
#     console: List[str] = []


# class ScidraWorkflow(CamelModel):
#     id: int
#     user_id: str
#     project_id: str
#     workflow_description: str = ""
#     jobs: List[ScidraJob] = []
#     last_run: Optional[WorkflowRun] = None


class WorkflowProject(CamelModel):
    name: str
    user_id: str
    gitlab_project_id: int
    created: Optional[str]
    timezone: str = "UTC"

    @validator("timezone")
    def must_be_valid_timezone(cls, v):
        if v not in pytz.all_timezones:
            raise ValueError(f"{v} is not a valid timezone")
        return v



import typing

import ujson
from starlette.responses import UJSONResponse as BaseUJSONResponse


class UJSONResponse(BaseUJSONResponse):
    """
    Custom Response class, based on default UJSONResponse, but with differences:
        * Allows to have forward slashes inside strings of JSON
        * Limits output to ASCII and escapes all extended characters above 127.

    Should be used as `response_class` argument to routes of your app:

    .. code-block:: python

        app = FastAPI()


        @app.get("/", response_class=UJSONResponse)
        async def root():
            return {"a": "b"}
    """

    def render(self, content: typing.Any) -> bytes:
        return ujson.dumps(
            content, ensure_ascii=True, escape_forward_slashes=False
        ).encode("utf-8")



from datetime import datetime
from typing import List, Optional

from pydantic import AnyUrl, EmailStr

from fa_common.models import CamelModel


class AuthUser(CamelModel):
    sub: str
    name: str = "Unknown User"
    email: Optional[EmailStr]
    nickname: Optional[str] = None
    email_verified: bool = False
    picture: Optional[AnyUrl] = None
    updated_at: Optional[datetime] = None
    scopes: List[str] = []
